#    Qlattice (https://github.com/jinluchang/qlattice)
#
#    Copyright (C) 2021
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import qlat_gpt as qg
import qlat as q
import gpt as g
import rbc_ukqcd as ru
import rbc_ukqcd_params as rup
import numpy as np
import pprint
import os

from auto_contractor.eval import *
from auto_contractor.operators import *

def get_save_path(fn):
    return os.path.join("results", fn)

def get_load_path(fn):
    if fn is None:
        return None
    path_list = [ "results" ]
    for path in path_list:
        p = os.path.join(path, fn)
        if q.does_file_exist_sync_node(p):
            return p
    return None

@q.timer
def compute_prop(inv, src, *, tag, sfw):
    sol = inv * src
    sol.save_double(sfw, tag)
    return sol

@q.timer
def compute_prop_psrc(gf, xg, job_tag, inv_type, inv_acc, *, idx, sfw, eig, finished_tags):
    tag = f"xg=({xg[0]},{xg[1]},{xg[2]},{xg[3]}) ; type={inv_type} ; accuracy={inv_acc}"
    if tag in finished_tags:
        return None
    q.check_stop()
    q.check_time_limit()
    q.displayln_info(f"compute_prop_psrc: idx={idx} xg={xg}", job_tag, inv_type, inv_acc)
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, eig = eig)
    total_site = ru.get_total_site(job_tag)
    geo = q.Geometry(total_site, 1)
    src = q.mk_point_src(geo, xg)
    compute_prop(inv, src, tag = tag, sfw = sfw)

def get_all_points(total_site, *, tslice = None):
    all_points = []
    if tslice is None:
        for t in range(total_site[3]):
            all_points += get_all_points(total_site, tslice = t)
        return all_points
    t = tslice
    for x in range(total_site[0]):
        for y in range(total_site[1]):
            for z in range(total_site[2]):
                all_points.append([x, y, z, t,])
    return all_points

@q.timer
def compute_prop_psrc_all(gf, job_tag, inv_type, *, path_s, eig):
    finished_tags = q.properly_truncate_fields(get_save_path(path_s + ".acc"))
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", [ 1, 1, 1, 2 ])
    total_site = ru.get_total_site(job_tag)
    inv_acc = 2
    for idx, xg in enumerate(get_all_points(total_site)):
        compute_prop_psrc(gf, xg, job_tag, inv_type, inv_acc,
                idx = idx, sfw = sfw, eig = eig,
                finished_tags = finished_tags)
    q.clean_cache(q.cache_inv)
    sfw.close()
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))

@q.timer
def compute_prop_wsrc(gf, gt, tslice, job_tag, inv_type, inv_acc, *, idx, sfw, path_sp, eig, finished_tags):
    tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in finished_tags:
        return None
    q.check_stop()
    q.check_time_limit()
    q.displayln_info(f"compute_prop_wsrc: idx={idx} tslice={tslice}", job_tag, inv_type, inv_acc)
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, gt = gt, eig = eig)
    total_site = ru.get_total_site(job_tag)
    geo = q.Geometry(total_site, 1)
    src = q.mk_wall_src(geo, tslice)
    prop = compute_prop(inv, src, tag = tag, sfw = sfw)
    fn_spw = os.path.join(path_sp, f"{tag} ; wsnk.lat")
    prop.glb_sum_tslice().save(get_save_path(fn_spw))

def get_all_walls(time_vol):
    all_walls = []
    for t in range(time_vol):
        all_walls.append(t)
    return all_walls

@q.timer
def compute_prop_wsrc_all(gf, gt, job_tag, inv_type, *, path_s, path_sp, eig):
    finished_tags = q.properly_truncate_fields(get_save_path(path_s + ".acc"))
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", [ 1, 1, 1, 2 ])
    total_site = ru.get_total_site(job_tag)
    inv_acc = 2
    for idx, tslice in enumerate(get_all_walls(total_site[3])):
        compute_prop_wsrc(gf, gt, tslice, job_tag, inv_type, inv_acc,
                idx = idx, sfw = sfw, path_sp = path_sp, eig = eig,
                finished_tags = finished_tags)
    q.clean_cache(q.cache_inv)
    sfw.close()
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))

@q.timer
def load_prop_psrc_all(job_tag, traj, flavor : str, path_s : str, psel):
    cache = q.mk_cache(f"prop_cache-{job_tag}-{traj}", flavor)
    total_site = ru.get_total_site(job_tag)
    if flavor in ["l", "u", "d",]:
        inv_type = 0
    elif flavor in ["s",]:
        inv_type = 1
    elif flavor in ["c",]:
        inv_type = 2
    else:
        assert False
    inv_acc = 2
    sfr = q.open_fields(get_load_path(path_s), "r")
    for idx, xg in enumerate(get_all_points(total_site)):
        q.displayln_info(f"load_prop_psrc_all: idx={idx} xg={xg} flavor={flavor} path_s={path_s}")
        prop = q.Prop()
        tag = f"xg=({xg[0]},{xg[1]},{xg[2]},{xg[3]}) ; type={inv_type} ; accuracy={inv_acc}"
        prop.load_double(sfr, tag)
        # convert to GPT/Grid prop mspincolor order
        prop_msc = q.convert_mspincolor_from_wm(prop)
        sp_prop_msc = q.PselProp(psel)
        sp_prop_msc @= prop_msc
        cache[f"xg=({xg[0]},{xg[1]},{xg[2]},{xg[3]})"] = sp_prop_msc
    sfr.close()

@q.timer
def load_prop_wsrc_all(job_tag, traj, flavor : str, path_s : str, path_sp : str, psel, gt):
    cache = q.mk_cache(f"prop_cache-{job_tag}-{traj}", flavor)
    total_site = ru.get_total_site(job_tag)
    psel_ts = q.get_psel_tslice(total_site)
    if flavor in ["l", "u", "d",]:
        inv_type = 0
    elif flavor in ["s",]:
        inv_type = 1
    elif flavor in ["c",]:
        inv_type = 2
    else:
        assert False
    inv_acc = 2
    gt_inv = gt.inv()
    sfr = q.open_fields(get_load_path(path_s), "r")
    for idx, tslice in enumerate(get_all_walls(total_site[3])):
        q.displayln_info(f"load_prop_wsrc_all: idx={idx} tslice={tslice} flavor={flavor} path_s={path_s} path_sp={path_sp}")
        prop = q.Prop()
        tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
        prop.load_double(sfr, tag)
        prop = gt_inv * prop
        # convert to GPT/Grid prop mspincolor order
        prop_msc = q.convert_mspincolor_from_wm(prop)
        sp_prop_msc = q.PselProp(psel)
        sp_prop_msc @= prop_msc
        cache[f"tslice={tslice}"] = sp_prop_msc
        # load wsnk prop
        fn_spw = os.path.join(path_sp, f"{tag} ; wsnk.lat")
        spw_prop = q.PselProp(psel_ts)
        spw_prop.load(get_load_path(fn_spw))
        spw_prop_msc = q.convert_mspincolor_from_wm(spw_prop)
        cache[f"tslice={tslice} ; wsnk"] = spw_prop_msc
    sfr.close()

@q.timer
def get_prop_psrc(prop_cache, flavor : str, xg_src):
    # prop_cache[flavor][src_p] = prop
    # call load_prop_psrc_all(flavor, path_s) first
    return prop_cache[flavor][f"xg=({xg_src[0]},{xg_src[1]},{xg_src[2]},{xg_src[3]})"]

@q.timer
def get_prop_wsrc(prop_cache, flavor : str, t_src):
    # prop_cache[flavor][src_p] = prop
    # call load_prop_wsrc_all(flavor, path_s) first
    return prop_cache[flavor][f"tslice={t_src}"]

@q.timer
def get_prop_wsnk_wsrc(prop_cache, flavor : str, t_snk, t_src):
    # prop_cache[flavor][src_p] = prop
    # call load_prop_wsrc_all(flavor, path_s) first
    return prop_cache[flavor][f"tslice={t_src} ; wsnk"].get_elem(t_snk)

@q.timer
def get_prop_psnk_psrc(prop_cache, flavor : str, p_snk, p_src, *, psel_pos_dict):
    if isinstance(p_snk, list) and isinstance(p_src, list):
        assert 4 == len(p_snk)
        assert 4 == len(p_src)
        xg_snk = p_snk
        xg_src = p_src
        msc = get_prop_psrc(prop_cache, flavor, xg_src).get_elem(psel_pos_dict[tuple(xg_snk)])
    else:
        assert isinstance(p_snk, tuple) and isinstance(p_src, tuple)
        assert 2 == len(p_snk)
        assert 2 == len(p_src)
        type_snk, pos_snk = p_snk
        type_src, pos_src = p_src
        if type_snk == "wall" and type_src == "wall":
            msc = get_prop_wsnk_wsrc(prop_cache, flavor, pos_snk, pos_src)
        elif type_snk == "point" and type_src == "wall":
            msc = get_prop_wsrc(prop_cache, flavor, pos_src).get_elem(psel_pos_dict[tuple(pos_snk)])
        elif type_snk == "wall" and type_src == "point":
            msc = ascontiguoustensor(
                    ascontiguoustensor(
                        g.gamma[5]
                        * adj_msc(
                            as_mspincolor(
                                get_prop_wsrc(prop_cache, flavor, pos_snk).get_elem(psel_pos_dict[tuple(pos_src)]))))
                    * g.gamma[5])
        elif type_snk == "point" and type_src == "point":
            msc = get_prop_psrc(prop_cache, flavor, pos_src).get_elem(psel_pos_dict[tuple(pos_snk)])
        else:
            raise Exception("get_prop_psnk_psrc unknown p_snk={p_snk} p_src={p_src}")
    return as_mspincolor(msc)

def mk_get_prop(prop_cache, *, psel_pos_dict):
    def get_prop(flavor, p_snk, p_src):
        return get_prop_psnk_psrc(prop_cache, flavor, p_snk, p_src, psel_pos_dict = psel_pos_dict)
    return get_prop

@q.timer
def auto_contractor_simple_test(job_tag, traj, get_prop):
    q.displayln_info(g.gamma[5] * get_prop("l", [1, 2, 3, 2], [1, 2, 3, 2]))
    q.displayln_info(g.trace(g.gamma[5] * get_prop("l", [1, 2, 3, 2], [1, 2, 3, 2])))
    q.displayln_info(g.gamma[5] * get_prop("l", [1, 2, 3, 2], [1, 2, 3, 2]) - get_prop("l", [1, 2, 3, 2], [1, 2, 3, 2]))
    q.displayln_info(g.gamma[5] * adj_msc(get_prop("l", [1, 2, 3, 2], [1, 2, 3, 2])))
    q.displayln_info(g.norm2(g.gamma[5] * adj_msc(get_prop("l", [1, 2, 3, 2], [1, 2, 3, 4])) * g.gamma[5] - get_prop("l", [1, 2, 3, 4], [1, 2, 3, 2])))
    expr = (1
            * Qb("d", "x1", "s1", "c1")
            * G(5, "s1", "s2")
            * Qv("u", "x1", "s2", "c1")
            * Qb("u", "x2", "s3", "c2")
            * G(5, "s3", "s4")
            * Qv("d", "x2", "s4", "c2"))
    expr = contract_expr(expr)
    expr.simplify(is_isospin_symmetric_limit = True)
    cexpr = mk_cexpr(expr)
    cexpr.collect_op()
    q.displayln_info(cexpr)
    positions_dict = {}
    positions_dict["x1"] = [1, 2, 3, 4]
    positions_dict["x2"] = [1, 2, 3, 2]
    val = eval_cexpr(cexpr, positions_dict = positions_dict, get_prop = get_prop, is_only_total = "total")
    q.displayln_info("eval_cexpr: ", val)
    q.displayln_info("gpt_direct: ",
            -g.trace(
                g.gamma[5]
                * get_prop("l", [1, 2, 3, 4], [1, 2, 3, 2])
                * g.gamma[5]
                * get_prop("l", [1, 2, 3, 2], [1, 2, 3, 4])
                ))

@q.timer
def get_cexpr_vev():
    def calc_cexpr():
        s = new_spin_index()
        c = new_color_index()
        p = "x"
        exprs = [
                Qb("u", "x", s, c) * Qv("u", "x", s, c) + "u_bar*u",
                Qb("d", "x", s, c) * Qv("d", "x", s, c) + "d_bar*d",
                Qb("s", "x", s, c) * Qv("s", "x", s, c) + "s_bar*s",
                Qb("c", "x", s, c) * Qv("c", "x", s, c) + "c_bar*c",
                ]
        cexpr = contract_simplify_compile(*exprs, is_isospin_symmetric_limit = True)
        q.displayln_info(display_cexpr(cexpr))
        cexpr.collect_op()
        return cexpr
    cexpr = q.pickle_cache_call(calc_cexpr, f"cache/auto_contractor_cexpr/vev-cexpr.pickle")
    q.displayln_info(display_cexpr(cexpr))
    return cexpr

@q.timer
def auto_contractor_vev(job_tag, traj, get_prop, num_trials):
    total_site = ru.get_total_site(job_tag)
    vol = total_site[0] * total_site[1] * total_site[2]
    cexpr = get_cexpr_vev()
    rng_state = q.RngState("seed")
    def positions_dict_maker(idx):
        rs = rng_state.split(str(idx))
        x = rs.c_rand_gen(total_site)
        pd = {
                "x" : ("point", x,),
                }
        facs = [ 1.0, ]
        return pd, facs
    names_fac = [ "rest", ]
    trial_indices = get_mpi_chunk(range(num_trials))
    results_list = eval_cexpr_simulation(
            cexpr,
            positions_dict_maker = positions_dict_maker,
            trial_indices = trial_indices,
            get_prop = get_prop,
            is_only_total = "total")
    for name_fac, results in zip(names_fac, results_list):
        q.displayln_info(f"{name_fac} :")
        for k, v in results.items():
            q.displayln_info(f"{name_fac} {k}:\n  {v}")

@q.timer
def auto_contractor_meson_corr(job_tag, traj, get_prop, num_trials):
    total_site = ru.get_total_site(job_tag)
    vol = total_site[0] * total_site[1] * total_site[2]
    exprs = [
            vol**2 * mk_pi_p("x2", True) * mk_pi_p("x1"),
            vol**2 * mk_k_p("x2", True) * mk_k_p("x1"),
            vol**2 * mk_k_m("x2", True) * mk_k_m("x1"),
            vol * mk_pi_p("x2", True) * mk_pi_p("t1"),
            vol * mk_pi_p("t2", True) * mk_pi_p("x1"),
            mk_pi_p("t2", True) * mk_pi_p("t1"),
            ]
    cexpr = contract_simplify_compile(*exprs, is_isospin_symmetric_limit = True)
    q.displayln_info(display_cexpr(cexpr))
    cexpr.collect_op()
    q.displayln_info(display_cexpr(cexpr))
    rng_state = q.RngState("seed")
    def positions_dict_maker(idx):
        rs = rng_state.split(str(idx))
        t2 = 5
        x1 = rs.c_rand_gen(total_site)
        x2 = rs.c_rand_gen(total_site)
        x2[3] = (x1[3] + t2) % total_site[3]
        pd = {
                "x1" : ("point", x1,),
                "x2" : ("point", x2,),
                "t1" : ("wall", x1[3],),
                "t2" : ("wall", x2[3],),
                }
        lmom = [ 2 * math.pi / total_site[i] for i in range(3) ]
        facs = [
                1.0,
                sum([ cmath.rect(1.0, (x2[i] - x1[i]) * lmom[i]).real for i in range(3) ]) / 3.0,
                sum([ cmath.rect(1.0, (x2[i] - x1[i]) * 2 * lmom[i]).real for i in range(3) ]) / 3.0,
                ]
        return pd, facs
    names_fac = ["rest", "mom1", "mom2",]
    trial_indices = get_mpi_chunk(range(num_trials))
    results_list = eval_cexpr_simulation(cexpr, positions_dict_maker = positions_dict_maker, trial_indices = trial_indices, get_prop = get_prop, is_only_total = "typed_total")
    for name_fac, results in zip(names_fac, results_list):
        q.displayln_info(f"{name_fac} :")
        for k, v in results.items():
            q.displayln_info(f"{name_fac} {k}:\n  {v}")

@q.timer
def get_cexpr_test_corr(vol):
    def calc_cexpr():
        exprs = [
                vol**2 * mk_k0pi0("x2_1", "x2_2", True) * mk_k0pi0("x1_1", "x1_2"),
                vol**2 * mk_k0pi0("x2_1", "x2_2", True) * mk_kmpip("x1_1", "x1_2"),
                vol**2 * mk_kmpip("x2_1", "x2_2", True) * mk_kmpip("x1_1", "x1_2"),
                vol**2 * mk_kppim("x2_1", "x2_2", True) * mk_kppim("x1_1", "x1_2"),
                vol**2 * mk_k0barpi0("x2_1", "x2_2", True) * mk_k0barpi0("x1_1", "x1_2"),
                vol**2 * mk_kpi_0_i1half("x2_1", "x2_2", True) * mk_kpi_0_i1half("x1_1", "x1_2"),
                vol**2 * mk_kpi_0_i1half("x2_1", "x2_2", True) * mk_kpi_m_i1half("x1_1", "x1_2"),
                vol**2 * mk_kpi_m_i1half("x2_1", "x2_2", True) * mk_kpi_m_i1half("x1_1", "x1_2"),
                vol**2 * mk_kpi_0_i1half("x2_1", "x2_2", True) * mk_kpi_0_i3halves("x1_1", "x1_2"),
                vol**2 * mk_kpi_m_i1half("x2_1", "x2_2", True) * mk_kpi_m1_i3halves("x1_1", "x1_2"),
                vol**2 * mk_kpi_0_i3halves("x2_1", "x2_2", True) * mk_kpi_0_i3halves("x1_1", "x1_2"),
                vol**2 * mk_k_0("x2_1", True) * mk_kpi_0_i1half("x1_1", "x1_2"),
                vol**2 * mk_k_0("x2_1", True) * mk_kpi_0_i3halves("x1_1", "x1_2"),
                vol**2 * mk_k_m("x2_1", True) * mk_kpi_m_i1half("x1_1", "x1_2"),
                vol**2 * mk_k_m("x2_1", True) * mk_kpi_m1_i3halves("x1_1", "x1_2"),
                vol**2 * mk_kpi_p_i3halves("x2_1", "x2_2", True) * mk_kpi_0_i3halves("x1_1", "x1_2"),
                vol**2 * mk_kpi_m1_i3halves("x2_1", "x2_2", True) * mk_kpi_0_i3halves("x1_1", "x1_2"),
                vol**2 * mk_kpi_m1_i3halves("x2_1", "x2_2", True) * mk_kpi_m1_i3halves("x1_1", "x1_2"),
                vol**2 * mk_kpi_m2_i3halves("x2_1", "x2_2", True) * mk_kpi_0_i3halves("x1_1", "x1_2"),
                vol**2 * mk_kpi_m2_i3halves("x2_1", "x2_2", True) * mk_kpi_m2_i3halves("x1_1", "x1_2"),
                ]
        cexpr = contract_simplify_compile(*exprs, is_isospin_symmetric_limit = True)
        q.displayln_info(display_cexpr(cexpr))
        cexpr.collect_op()
        return cexpr
    cexpr = q.pickle_cache_call(calc_cexpr, f"cache/auto_contractor_cexpr/test_corr-cexpr.{vol}.pickle")
    q.displayln_info(display_cexpr(cexpr))
    return cexpr

@q.timer
def auto_contractor_test_corr(job_tag, traj, get_prop, num_trials):
    total_site = ru.get_total_site(job_tag)
    vol = total_site[0] * total_site[1] * total_site[2]
    cexpr = get_cexpr_test_corr(vol)
    rng_state = q.RngState("seed")
    def positions_dict_maker(idx):
        rs = rng_state.split(str(idx))
        t1_2 = 2
        t2_1 = 5
        t2_2 = 7
        x1_1 = rs.c_rand_gen(total_site)
        x1_2 = rs.c_rand_gen(total_site)
        x2_1 = rs.c_rand_gen(total_site)
        x2_2 = rs.c_rand_gen(total_site)
        x1_2[3] = (x1_1[3] + t1_2) % total_site[3]
        x2_1[3] = (x1_1[3] + t2_1) % total_site[3]
        x2_2[3] = (x1_1[3] + t2_2) % total_site[3]
        pd = {
                "x1_1" : ("point", x1_1,),
                "x1_2" : ("point", x1_2,),
                "x2_1" : ("point", x2_1,),
                "x2_2" : ("point", x2_2,),
                }
        lmom = [ 2 * math.pi / total_site[i] for i in range(3) ]
        fac1 = sum([ 2 * cmath.rect(1.0, (x1_1[i] - x1_2[i]) * lmom[i]).real for i in range(3) ]) / np.sqrt(6)
        fac2 = sum([ 2 * cmath.rect(1.0, (x2_1[i] - x2_2[i]) * lmom[i]).real for i in range(3) ]) / np.sqrt(6)
        facs = [1.0, fac1, fac2, fac1 * fac2,]
        return pd, facs
    names_fac = ["rest-rest", "rest-moving", "moving-rest", "moving-moving",]
    trial_indices = get_mpi_chunk(range(num_trials))
    results_list = eval_cexpr_simulation(cexpr, positions_dict_maker = positions_dict_maker, trial_indices = trial_indices, get_prop = get_prop, is_only_total = "typed_total")
    for name_fac, results in zip(names_fac, results_list):
        q.displayln_info(f"{name_fac} :")
        for k, v in results.items():
            q.displayln_info(f"{name_fac} {k}:\n  {v}")

@q.timer
def get_cexpr_pipi_corr(vol):
    def calc_cexpr():
        exprs = [
                vol**4 * mk_pipi_i22("x2_1", "x2_2", True) * mk_pipi_i22("x1_1", "x1_2"),
                vol**4 * mk_pipi_i11("x2_1", "x2_2", True) * mk_pipi_i11("x1_1", "x1_2"),
                vol**4 * mk_pipi_i0("x2_1", "x2_2", True) * mk_pipi_i0("x1_1", "x1_2"),
                vol**2 * mk_pipi_i0("x2_1", "x2_2", True),
                vol**2 * mk_pipi_i0("x1_1", "x1_2"),
                vol**2 * mk_sigma("x2_1", True) * mk_sigma("x1_1"),
                vol**2 * mk_sigma("x2_1", True) * mk_sigma("x1_2"),
                vol**2 * mk_sigma("x2_2", True) * mk_sigma("x1_1"),
                vol**2 * mk_sigma("x2_2", True) * mk_sigma("x1_2"),
                vol * mk_sigma("x1_1"),
                vol * mk_sigma("x1_2"),
                vol * mk_sigma("x2_1", True),
                vol * mk_sigma("x2_2", True),
                vol**3 * mk_sigma("x2_1", True) * mk_pipi_i0("x1_1", "x1_2"),
                vol**3 * mk_sigma("x2_2", True) * mk_pipi_i0("x1_1", "x1_2"),
                vol**3 * mk_pipi_i0("x2_1", "x2_2", True) * mk_sigma("x1_1"),
                vol**3 * mk_pipi_i0("x2_1", "x2_2", True) * mk_sigma("x1_2"),
                ]
        cexpr = contract_simplify_compile(*exprs, is_isospin_symmetric_limit = True)
        q.displayln_info(display_cexpr(cexpr))
        cexpr.collect_op()
        return cexpr
    cexpr = q.pickle_cache_call(calc_cexpr, f"cache/auto_contractor_cexpr/pipi_corr-cexpr.{vol}.pickle")
    q.displayln_info(display_cexpr(cexpr))
    return cexpr

@q.timer
def auto_contractor_pipi_corr(job_tag, traj, get_prop, num_trials):
    total_site = ru.get_total_site(job_tag)
    vol = total_site[0] * total_site[1] * total_site[2]
    cexpr = get_cexpr_pipi_corr(vol)
    rng_state = q.RngState("seed")
    def positions_dict_maker(idx):
        rs = rng_state.split(str(idx))
        t1_2 = 2
        t2_1 = 5
        t2_2 = 7
        x1_1 = rs.c_rand_gen(total_site)
        x1_2 = rs.c_rand_gen(total_site)
        x2_1 = rs.c_rand_gen(total_site)
        x2_2 = rs.c_rand_gen(total_site)
        x1_2[3] = (x1_1[3] + t1_2) % total_site[3]
        x2_1[3] = (x1_1[3] + t2_1) % total_site[3]
        x2_2[3] = (x1_1[3] + t2_2) % total_site[3]
        pd = {
                "x1_1" : ("point", x1_1,),
                "x1_2" : ("point", x1_2,),
                "x2_1" : ("point", x2_1,),
                "x2_2" : ("point", x2_2,),
                }
        lmom = [ 2 * math.pi / total_site[i] for i in range(3) ]
        fac1 = sum([ 2 * cmath.rect(1.0, (x1_1[i] - x1_2[i]) * lmom[i]).real for i in range(3) ]) / np.sqrt(6)
        fac2 = sum([ 2 * cmath.rect(1.0, (x2_1[i] - x2_2[i]) * lmom[i]).real for i in range(3) ]) / np.sqrt(6)
        facs = [1.0, fac1, fac2, fac1 * fac2,]
        return pd, facs
    names_fac = ["rest-rest", "rest-moving", "moving-rest", "moving-moving",]
    trial_indices = get_mpi_chunk(range(num_trials))
    results_list = eval_cexpr_simulation(cexpr, positions_dict_maker = positions_dict_maker, trial_indices = trial_indices, get_prop = get_prop, is_only_total = "typed_total")
    for name_fac, results in zip(names_fac, results_list):
        q.displayln_info(f"{name_fac} :")
        for k, v in results.items():
            q.displayln_info(f"{name_fac} {k}:\n  {v}")

@q.timer
def get_cexpr_kpipi(vol):
    @q.timer
    def calc_cexpr():
        total_site = ru.get_total_site(job_tag)
        vol = total_site[0] * total_site[1] * total_site[2]
        exprs_odd_ops = [
                vol * mk_Q1("x", "odd") + "Q1(o)",
                vol * mk_Q2("x", "odd") + "Q2(o)",
                vol * mk_Q3("x", "odd") + "Q3(o)",
                vol * mk_Q4("x", "odd") + "Q4(o)",
                vol * mk_Q5("x", "odd") + "Q5(o)",
                vol * mk_Q6("x", "odd") + "Q6(o)",
                vol * mk_Q7("x", "odd") + "Q7(o)",
                vol * mk_Q8("x", "odd") + "Q8(o)",
                vol * mk_Q9("x", "odd") + "Q9(o)",
                vol * mk_Q10("x", "odd") + "Q10(o)",
                vol * mk_Qsub("x", "odd") + "Qs(o)",
                ]
        exprs_even_ops = [
                vol * mk_Q1("x", "even") + "Q1(e)",
                vol * mk_Q2("x", "even") + "Q2(e)",
                vol * mk_Q3("x", "even") + "Q3(e)",
                vol * mk_Q4("x", "even") + "Q4(e)",
                vol * mk_Q5("x", "even") + "Q5(e)",
                vol * mk_Q6("x", "even") + "Q6(e)",
                vol * mk_Q7("x", "even") + "Q7(e)",
                vol * mk_Q8("x", "even") + "Q8(e)",
                vol * mk_Q9("x", "even") + "Q9(e)",
                vol * mk_Q10("x", "even") + "Q10(e)",
                vol * mk_Qsub("x", "even") + "Qs(e)",
                ]
        exprs_ops = exprs_odd_ops + exprs_even_ops
        exprs_k = [
                vol * mk_k_0("x2") + "K0",
                ]
        exprs_pipi = [
                vol**2 * mk_pipi_i0("x1_1", "x1_2", True) + "pipi_I0",
                vol**2 * mk_pipi_i20("x1_1", "x1_2", True) + "pipi_I2",
                vol * mk_sigma("x1_1", True) + "sigma_1",
                vol * mk_sigma("x1_2", True) + "sigma_2",
                vol * mk_pi_0("x1_1", True) + "pi0_1",
                vol * mk_pi_0("x1_2", True) + "pi0_2",
                mk_expr(1) + "1",
                ]
        exprs = []
        for expr_k in exprs_k:
            for expr_pipi in exprs_pipi:
                for expr_op in exprs_ops:
                    exprs.append(expr_pipi * expr_op * expr_k)
        diagram_type_dict = dict()
        diagram_type_dict[((('x', 'x'), 1), (('x', 'x1_1'), 1), (('x1_1', 'x1_2'), 1), (('x1_2', 'x2'), 1), (('x2', 'x'), 1))] = "Type3"
        diagram_type_dict[((('x', 'x'), 1), (('x', 'x2'), 1), (('x1_1', 'x1_2'), 1), (('x1_2', 'x1_1'), 1), (('x2', 'x'), 1))] = "Type4"
        diagram_type_dict[((('x', 'x1_1'), 1), (('x', 'x1_2'), 1), (('x1_1', 'x'), 1), (('x1_2', 'x2'), 1), (('x2', 'x'), 1))] = "Type1"
        diagram_type_dict[((('x', 'x1_1'), 1), (('x', 'x2'), 1), (('x1_1', 'x1_2'), 1), (('x1_2', 'x'), 1), (('x2', 'x'), 1))] = "Type2"
        diagram_type_dict[((('x', 'x1_1'), 1), (('x1_1', 'x1_2'), 1), (('x1_2', 'x2'), 1), (('x2', 'x'), 1))] = "Type3"
        diagram_type_dict[((('x', 'x2'), 1), (('x1_1', 'x1_2'), 1), (('x1_2', 'x1_1'), 1), (('x2', 'x'), 1))] = "Type4"
        diagram_type_dict[((('x', 'x'), 1), (('x', 'x1_1'), 1), (('x1_1', 'x2'), 1), (('x2', 'x'), 1))] = "Type3"
        diagram_type_dict[((('x', 'x'), 1), (('x', 'x2'), 1), (('x1_1', 'x1_1'), 1), (('x2', 'x'), 1))] = "Type4"
        diagram_type_dict[((('x', 'x1_1'), 1), (('x', 'x2'), 1), (('x1_1', 'x'), 1), (('x2', 'x'), 1))] = "Type2"
        diagram_type_dict[((('x', 'x1_1'), 1), (('x1_1', 'x2'), 1), (('x2', 'x'), 1))] = "Type3"
        diagram_type_dict[((('x', 'x2'), 1), (('x1_1', 'x1_1'), 1), (('x2', 'x'), 1))] = "Type4"
        diagram_type_dict[((('x', 'x'), 1), (('x', 'x1_2'), 1), (('x1_2', 'x2'), 1), (('x2', 'x'), 1))] = "Type3"
        diagram_type_dict[((('x', 'x'), 1), (('x', 'x2'), 1), (('x1_2', 'x1_2'), 1), (('x2', 'x'), 1))] = "Type4"
        diagram_type_dict[((('x', 'x1_2'), 1), (('x', 'x2'), 1), (('x1_2', 'x'), 1), (('x2', 'x'), 1))] = "Type2"
        diagram_type_dict[((('x', 'x1_2'), 1), (('x1_2', 'x2'), 1), (('x2', 'x'), 1))] = "Type3"
        diagram_type_dict[((('x', 'x2'), 1), (('x1_2', 'x1_2'), 1), (('x2', 'x'), 1))] = "Type4"
        diagram_type_dict[((('x', 'x'), 1), (('x', 'x2'), 1), (('x2', 'x'), 1))] = "Type4"
        diagram_type_dict[((('x', 'x2'), 1), (('x2', 'x'), 1))] = "Type4"
        cexpr = contract_simplify_compile(*exprs, is_isospin_symmetric_limit = True, diagram_type_dict = diagram_type_dict)
        q.displayln_info(display_cexpr(cexpr))
        cexpr.collect_op()
        return cexpr
    cexpr = q.pickle_cache_call(calc_cexpr, f"cache/auto_contractor_cexpr/kpipi-cexpr.{vol}.pickle")
    q.displayln_info(display_cexpr(cexpr))
    return cexpr

@q.timer
def auto_contractor_kpipi_corr(job_tag, traj, get_prop, num_trials):
    total_site = ru.get_total_site(job_tag)
    vol = total_site[0] * total_site[1] * total_site[2]
    cexpr = get_cexpr_kpipi(vol)
    rng_state = q.RngState("seed")
    def positions_dict_maker(idx):
        rs = rng_state.split(str(idx))
        t1_1 = 5
        t1_2 = 7
        t = 3
        x1_1 = rs.c_rand_gen(total_site)
        x1_2 = rs.c_rand_gen(total_site)
        x = rs.c_rand_gen(total_site)
        x2 = rs.c_rand_gen(total_site)
        x1_1[3] = (x2[3] + t1_1) % total_site[3]
        x1_2[3] = (x2[3] + t1_2) % total_site[3]
        x[3] = (x2[3] + t) % total_site[3]
        pd = {
                "x1_1" : ("point", x1_1,),
                "x1_2" : ("point", x1_2,),
                "x" : ("point", x,),
                "x2" : ("point", x2,),
                }
        lmom = [ 2 * math.pi / total_site[i] for i in range(3) ]
        fac1 = sum([ 2 * cmath.rect(1.0, (x1_1[i] - x1_2[i]) * lmom[i]).real for i in range(3) ]) / np.sqrt(6)
        facs = [1.0, fac1]
        return pd, facs
    names_fac = [ "rest", "moving", ]
    trial_indices = get_mpi_chunk(range(num_trials))
    results_list = eval_cexpr_simulation(cexpr, positions_dict_maker = positions_dict_maker, trial_indices = trial_indices, get_prop = get_prop, is_only_total = "typed_total")
    q.qremove_all_info("analysis/kpipi")
    def mk_fn(info):
        def f(c):
            if c in "()<>/* ":
                return "_"
            else:
                return c
        info = "".join(map(f, info))
        while True:
            fn = info.replace("__", "_")
            if fn == info:
                break
            info = fn
        if fn[-1] == "_":
            fn = fn[:-1]
        return fn
    for name_fac, results in zip(names_fac, results_list):
        q.displayln_info(f"{name_fac} :")
        for k, v in results.items():
            q.displayln_info(f"{name_fac} {k}:\n  {v}")
            fn = "analysis/kpipi/" + mk_fn(f"{name_fac} {k}") + ".txt"
            [ a, e, ] = v
            q.qtouch(fn, f"0 {a.real} {a.imag} {e.real} {e.imag}\n")

@q.timer
def auto_contractor_kpipi_corr_81oprs(job_tag, traj, get_prop, num_trials):
    total_site = ru.get_total_site(job_tag)
    vol = total_site[0] * total_site[1] * total_site[2]
    exprs_odd_ops = [
            vol * mk_Q0_b81("x", "odd") + "Q0_b81(o)",
            vol * mk_Q1_b81("x", "odd") + "Q1_b81(o)",
            vol * mk_Q2_b81("x", "odd") + "Q2_b81(o)",
            vol * mk_Q3_b81("x", "odd") + "Q3_b81(o)",
            vol * mk_Q4_b81("x", "odd") + "Q4_b81(o)",
            vol * mk_Q5_b81("x", "odd") + "Q5_b81(o)",
            vol * mk_Q6_b81("x", "odd") + "Q6_b81(o)",
            vol * mk_Q7_b81("x", "odd") + "Q7_b81(o)",
            vol * mk_Q8_b81("x", "odd") + "Q8_b81(o)",
            ]
    exprs_even_ops = [
            vol * mk_Q0_b81("x", "even") + "Q0_b81(e)",
            vol * mk_Q1_b81("x", "even") + "Q1_b81(e)",
            vol * mk_Q2_b81("x", "even") + "Q2_b81(e)",
            vol * mk_Q3_b81("x", "even") + "Q3_b81(e)",
            vol * mk_Q4_b81("x", "even") + "Q4_b81(e)",
            vol * mk_Q5_b81("x", "even") + "Q5_b81(e)",
            vol * mk_Q6_b81("x", "even") + "Q6_b81(e)",
            vol * mk_Q7_b81("x", "even") + "Q7_b81(e)",
            vol * mk_Q8_b81("x", "even") + "Q8_b81(e)",
            ]
    exprs_ops = exprs_odd_ops + exprs_even_ops
    exprs_k = [
            vol * mk_k_0("x2") + "K0",
            ]
    exprs_pipi = [
            #vol**2 * mk_pipi_i0("x1_1", "x1_2", True) + "pipi_I0",
            vol**2 * mk_pipi_i20("x1_1", "x1_2", True) + "pipi_I2",
            #vol * mk_sigma("x1_1", True) + "sigma_1",
            #vol * mk_sigma("x1_2", True) + "sigma_2",
            #vol * mk_pi_0("x1_1", True) + "pi0_1",
            #vol * mk_pi_0("x1_2", True) + "pi0_2",
            #mk_expr(1) + "1",
            ]
    exprs = []
    for expr_k in exprs_k:
        for expr_pipi in exprs_pipi:
            for expr_op in exprs_ops:
                exprs.append(expr_pipi * expr_op * expr_k)
    diagram_type_dict = dict()
    cexpr = contract_simplify_compile(*exprs, is_isospin_symmetric_limit = True, diagram_type_dict = diagram_type_dict)
    q.displayln_info(display_cexpr(cexpr))
    cexpr.collect_op()
    q.displayln_info(display_cexpr(cexpr))
    rng_state = q.RngState("seed")
    def positions_dict_maker(idx):
        rs = rng_state.split(str(idx))
        t1_1 = 5
        t1_2 = 7
        t = 3
        x1_1 = rs.c_rand_gen(total_site)
        x1_2 = rs.c_rand_gen(total_site)
        x = rs.c_rand_gen(total_site)
        x2 = rs.c_rand_gen(total_site)
        x1_1[3] = (x2[3] + t1_1) % total_site[3]
        x1_2[3] = (x2[3] + t1_2) % total_site[3]
        x[3] = (x2[3] + t) % total_site[3]
        pd = {
                "x1_1" : ("point", x1_1,),
                "x1_2" : ("point", x1_2,),
                "x" : ("point", x,),
                "x2" : ("point", x2,),
                }
        lmom = [ 2 * math.pi / total_site[i] for i in range(3) ]
        fac1 = sum([ 2 * cmath.rect(1.0, (x1_1[i] - x1_2[i]) * lmom[i]).real for i in range(3) ]) / np.sqrt(6)
        facs = [1.0, fac1]
        return pd, facs
    names_fac = [ "rest", "moving", ]
    trial_indices = get_mpi_chunk(range(num_trials))
    results_list = eval_cexpr_simulation(cexpr, positions_dict_maker = positions_dict_maker, trial_indices = trial_indices, get_prop = get_prop, is_only_total = "typed_total")
    q.qmkdir_info("analysis")
    q.qremove_all_info("analysis/kpipi_b81")
    q.qmkdir_info("analysis/kpipi_b81")
    def mk_fn(info):
        def f(c):
            if c in "()<>/* ":
                return "_"
            else:
                return c
        info = "".join(map(f, info))
        while True:
            fn = info.replace("__", "_")
            if fn == info:
                break
            info = fn
        if fn[-1] == "_":
            fn = fn[:-1]
        return fn
    for name_fac, results in zip(names_fac, results_list):
        q.displayln_info(f"{name_fac} :")
        for k, v in results.items():
            q.displayln_info(f"{name_fac} {k}:\n  {v}")
            fn = "analysis/kpipi_b81/" + mk_fn(f"{name_fac} {k}") + ".txt"
            [ a, e, ] = v
            q.qtouch(fn, f"0 {a.real} {a.imag} {e.real} {e.imag}\n")

@q.timer
def get_cexpr_3f4f(vol):
    @q.timer
    def calc_cexpr():
        total_site = ru.get_total_site(job_tag)
        vol = total_site[0] * total_site[1] * total_site[2]
        exprs_odd_ops = [
            vol * mk_Q0_b81("x", "odd") + "Q0_b81(o)",
            vol * mk_Q1_b81("x", "odd") + "Q1_b81(o)",
            vol * mk_Q2_b81("x", "odd") + "Q2_b81(o)",
            vol * mk_Q3_b81("x", "odd") + "Q3_b81(o)",
            vol * mk_Q4_b81("x", "odd") + "Q4_b81(o)",
            vol * mk_Q5_b81("x", "odd") + "Q5_b81(o)",
            vol * mk_Q6_b81("x", "odd") + "Q6_b81(o)",
            vol * mk_Q7_b81("x", "odd") + "Q7_b81(o)",
            vol * mk_Q8_b81("x", "odd") + "Q8_b81(o)",
        ]
        exprs_even_ops = [
            vol * mk_Q0_b81("x", "even") + "Q0_b81(e)",
            vol * mk_Q1_b81("x", "even") + "Q1_b81(e)",
            vol * mk_Q2_b81("x", "even") + "Q2_b81(e)",
            vol * mk_Q3_b81("x", "even") + "Q3_b81(e)",
            vol * mk_Q4_b81("x", "even") + "Q4_b81(e)",
            vol * mk_Q5_b81("x", "even") + "Q5_b81(e)",
            vol * mk_Q6_b81("x", "even") + "Q6_b81(e)",
            vol * mk_Q7_b81("x", "even") + "Q7_b81(e)",
            vol * mk_Q8_b81("x", "even") + "Q8_b81(e)",
        ]
        exprs_ops = exprs_odd_ops + exprs_even_ops
        exprs_src = [
            vol * mk_k_0("t2_1") + "K0",
            vol * mk_kpi_0_i1half("t2_1", "t2_2") + "Kpi_0_I1half",
            vol * mk_kpi_0_i3halves("t2_1", "t2_2") + "Kpi_0_I3halves",
        ]
        exprs_snk = [
            vol**2 * mk_pipi_i20("t1_1", "t1_2", True) + "pipi_I2",
            vol**2 * mk_pipi_i0("t1_1", "t1_2", True) + "pipi_I0",
            vol * mk_sigma("t1_1", True) + "sigma",
            vol * mk_pi_0("t1_1", True) + "pi0",
            mk_expr(1) + "1",
        ]
        #    exprs_src_snk = [
        #        vol**2 * mk_
        #        ]
        exprs = []
        for expr_src in exprs_src:
            for expr_snk in exprs_snk:
                for expr_op in exprs_ops:
                    exprs.append(expr_snk * expr_op * expr_src)
        diagram_type_dict = dict()
        cexpr = contract_simplify_compile(*exprs, is_isospin_symmetric_limit = True, diagram_type_dict = diagram_type_dict)
        q.displayln_info(display_cexpr(cexpr))
        cexpr.collect_op()
        return cexpr
    cexpr = q.pickle_cache_call(calc_cexpr, f"cache/auto_contractor_cexpr/3f4f-cexpr.{vol}.pickle")
    q.displayln_info(display_cexpr(cexpr))
    return cexpr

@q.timer
def auto_contractor_3f4f_matching(job_tag, traj, get_prop, num_trials):
    total_site = ru.get_total_site(job_tag)
    vol = total_site[0] * total_site[1] * total_site[2]
    cexpr = get_cexpr_3f4f(vol)
    rng_state = q.RngState("seed")
    src_snk_seps = [8,10,12,14,16]
    tsep_src = -1
    tsep_snk = 1
    q.qmkdir_info("analysis")
    q.qremove_all_info("analysis/3f4f_b81")
    q.qmkdir_info("analysis/3f4f_b81")
    def mk_key(info):
        def f(c):
            if c in "()<>/* ":
                return "_"
            else:
                return c
        info = "".join(map(f, info))
        while True:
            fn = info.replace("__", "_")
            if fn == info:
                break
            info = fn
        if fn[-1] == "_":
            fn = fn[:-1]
        return fn
    #
    for tsnk_tsrc in src_snk_seps:
        max_top_tsrc = int(tsnk_tsrc/2)
        min_top_tsrc = int(tsnk_tsrc/2)
        fn = f"analysis/3f4f_b81/tsnk_tsrc{tsnk_tsrc}_top_tsrc{min_top_tsrc}-{max_top_tsrc}.bin"
        with open(fn, mode='wb') as f:
            for top_tsrc in range(min_top_tsrc,max_top_tsrc+1):
                def positions_dict_maker(idx):
                    rs = rng_state.split(str(idx))
                    tsrc1_top = - top_tsrc
                    tsrc2_top = tsep_src  + tsrc1_top
                    tsnk1_top = tsnk_tsrc + tsrc1_top
                    tsnk2_top = tsep_snk  + tsnk1_top
                    dt1_2 = 4
                    x = rs.c_rand_gen(total_site)
                    t2_1 = ( tsrc1_top + x[3] + total_site[3] ) % total_site[3]
                    t2_2 = ( tsrc2_top + x[3] + total_site[3] ) % total_site[3]
                    t1_1 = ( tsnk1_top + x[3] + total_site[3] ) % total_site[3]
                    t1_2 = ( tsnk2_top + x[3] + total_site[3] ) % total_site[3]
                    pd = {
                        "t1_1" : ("wall", t1_1,),
                        "t1_2" : ("wall", t1_2,),
                        "x" : ("point", x,),
                        "t2_1" : ("wall", t2_1,),
                        "t2_2" : ("wall", t2_2,),
                    }
                    facs = [1.0]
                    return pd, facs
                trial_indices = get_mpi_chunk(range(num_trials))
                results_list = eval_cexpr_simulation(cexpr, positions_dict_maker = positions_dict_maker, trial_indices = trial_indices, get_prop = get_prop, is_only_total = "total")
                #for results in results_list:
                results = results_list[0]
                for k, v in results.items():
                    if v[1].real == 0:
                        ratio_real = None
                    else:
                        ratio_real = v[0].real / v[1].real
                    if v[1].imag == 0:
                        ratio_imag = None
                    else:
                        ratio_imag = v[0].imag / v[1].imag
                    q.displayln_info(f"{k}:\n  {v}, ({ratio_real}, {ratio_imag})")
                    ###
                    [ a, e, ] = v
                    f.write(a)
                    f.write(e)
                if top_tsrc == min_top_tsrc and tsnk_tsrc == src_snk_seps[0]:
                    metafn = f"analysis/3f4f_b81/meta.txt"
                    with open(metafn, mode='w') as metaf:
                        for k, v in results.items():
                            key = mk_key(f"{k}")
                            metaf.write(f"{key}\n")

@q.timer_verbose
def run_gf(job_tag, traj):
    path_gf = get_load_path(f"configs/{job_tag}/ckpoint_lat.{traj}")
    if path_gf is None:
        if job_tag[:5] == "test-":
            gf = ru.mk_sample_gauge_field(job_tag, f"{traj}")
            path_gf = get_save_path(f"configs/{job_tag}/ckpoint_lat.{traj}")
            # gf.save(path_gf)
            qg.save_gauge_field(gf, path_gf)
        else:
            assert False
    get_gf = ru.load_config_lazy(job_tag, path_gf)
    return get_gf

@q.timer_verbose
def run_gt(job_tag, traj, get_gf):
    if None in [ get_gf, ]:
        return None
    path_gt = get_load_path(f"gauge-transform/{job_tag}/traj={traj}.field")
    if path_gt is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-gauge_fix_coulomb"):
            gf = get_gf()
            gt = qg.gauge_fix_coulomb(gf)
            gt.save_double(get_save_path(f"gauge-transform/{job_tag}/traj={traj}.field"))
            q.release_lock()
            return lambda : gt
        else:
            return None
    else:
        @q.timer_verbose
        def load_gt():
            gt = q.GaugeTransform()
            gt.load_double(path_gt)
            # ADJUST ME
            # qg.check_gauge_fix_coulomb(get_gf(), gt)
            #
            return gt
        get_gt = q.lazy_call(load_gt)
    return get_gt

@q.timer
def compute_eig(gf, job_tag, inv_type = 0, inv_acc = 0, *, path = None):
    # return a function ``get_eig''
    # ``get_eig()'' return the ``eig''
    load_eig = ru.load_eig_lazy(get_load_path(path), job_tag)
    if load_eig is not None:
        return load_eig
    # evec, evals = ru.mk_eig(gf, job_tag, inv_type, inv_acc)
    basis, cevec, smoothed_evals = ru.mk_ceig(gf, job_tag, inv_type, inv_acc)
    eig = [ basis, cevec, smoothed_evals ]
    ru.save_ceig(get_save_path(path + ".partial"), eig, job_tag, inv_type, inv_acc);
    q.qrename_info(get_save_path(path + ".partial"), get_save_path(path))
    test_eig(gf, eig, job_tag, inv_type)
    def get_eig():
        return eig
    return get_eig

@q.timer
def test_eig(gf, eig, job_tag, inv_type):
    geo = gf.geo()
    src = q.FermionField4d(geo)
    q.displayln_info(f"src norm {src.qnorm()}")
    src.set_rand(q.RngState("test_eig:{id(inv)}"))
    sol_ref = ru.get_inv(gf, job_tag, inv_type, inv_acc = 2, eig = eig, eps = 1e-10, mpi_split = False, timer = False) * src
    q.displayln_info(f"sol_ref norm {sol_ref.qnorm()} with eig")
    for inv_acc in [0, 1, 2]:
        sol = ru.get_inv(gf, job_tag, inv_type, inv_acc, eig = eig, mpi_split = False, timer = False) * src
        sol -= sol_ref
        q.displayln_info(f"sol diff norm {sol.qnorm()} inv_acc={inv_acc} with eig")
        sol = ru.get_inv(gf, job_tag, inv_type, inv_acc, mpi_split = False, timer = False) * src
        sol -= sol_ref
        q.displayln_info(f"sol diff norm {sol.qnorm()} inv_acc={inv_acc} without eig")

@q.timer_verbose
def run_eig(job_tag, traj, get_gf):
    if None in [ get_gf, ]:
        return None
    get_eig = ru.load_eig_lazy(get_load_path(f"eig/{job_tag}/traj={traj}"), job_tag)
    if get_eig is None and get_gf is not None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-run-eig"):
            q.qmkdir_info(get_save_path(f"eig"))
            q.qmkdir_info(get_save_path(f"eig/{job_tag}"))
            get_eig = compute_eig(get_gf(), job_tag, inv_type = 0, path = f"eig/{job_tag}/traj={traj}")
            q.release_lock()
    return get_eig

@q.timer_verbose
def run_prop(job_tag, traj, get_gf, get_gt, get_eig):
    for inv_type in [0, 1, 2,]:
        if get_load_path(f"prop-wsrc-{inv_type}/{job_tag}/traj={traj}") is None:
            if inv_type == 0 and get_eig is None:
                continue
            if q.obtain_lock(f"locks/{job_tag}-{traj}-compute-prop-wsrc-all-{inv_type}"):
                q.qmkdir_info(get_save_path(f"prop-wsrc-{inv_type}"))
                q.qmkdir_info(get_save_path(f"prop-wsrc-{inv_type}/{job_tag}"))
                if inv_type == 0:
                    eig = get_eig()
                else:
                    eig = None
                compute_prop_wsrc_all(get_gf(), get_gt(), job_tag, inv_type,
                        path_s = f"prop-wsrc-{inv_type}/{job_tag}/traj={traj}",
                        path_sp = f"psel-prop-wsrc-{inv_type}/{job_tag}/traj={traj}",
                        eig = eig)
                q.release_lock()
        #
        if get_load_path(f"prop-psrc-{inv_type}/{job_tag}/traj={traj}") is None:
            if inv_type == 0 and get_eig is None:
                continue
            if q.obtain_lock(f"locks/{job_tag}-{traj}-compute-prop-psrc-all-{inv_type}"):
                q.qmkdir_info(get_save_path(f"prop-psrc-{inv_type}"))
                q.qmkdir_info(get_save_path(f"prop-psrc-{inv_type}/{job_tag}"))
                if inv_type == 0:
                    eig = get_eig()
                else:
                    eig = None
                compute_prop_psrc_all(get_gf(), job_tag, inv_type,
                        path_s = f"prop-psrc-{inv_type}/{job_tag}/traj={traj}",
                        eig = eig)
                q.release_lock()
    #
    path_prop_list = \
            [ get_load_path(f"prop-psrc-{inv_type}/{job_tag}/traj={traj}") for inv_type in [0, 1, 2,] ] + \
            [ get_load_path(f"prop-wsrc-{inv_type}/{job_tag}/traj={traj}") for inv_type in [0, 1, 2,] ]
    if all(map(lambda x : x is not None, path_prop_list)):
        total_site = ru.get_total_site(job_tag)
        psel = q.PointsSelection(get_all_points(total_site))
        psel_pos_dict = dict([ (tuple(pos), i) for i, pos in enumerate(psel.to_list()) ])
        load_prop_wsrc_all(job_tag, traj, "l",
                f"prop-wsrc-0/{job_tag}/traj={traj}",
                f"psel-prop-wsrc-0/{job_tag}/traj={traj}",
                psel, get_gt())
        load_prop_wsrc_all(job_tag, traj, "s",
                f"prop-wsrc-1/{job_tag}/traj={traj}",
                f"psel-prop-wsrc-1/{job_tag}/traj={traj}",
                psel, get_gt())
        load_prop_wsrc_all(job_tag, traj, "c",
                f"prop-wsrc-2/{job_tag}/traj={traj}",
                f"psel-prop-wsrc-2/{job_tag}/traj={traj}",
                psel, get_gt())
        load_prop_psrc_all(job_tag, traj, "l", f"prop-psrc-0/{job_tag}/traj={traj}", psel)
        load_prop_psrc_all(job_tag, traj, "s", f"prop-psrc-1/{job_tag}/traj={traj}", psel)
        load_prop_psrc_all(job_tag, traj, "c", f"prop-psrc-2/{job_tag}/traj={traj}", psel)
        prop_cache = q.mk_cache(f"prop_cache-{job_tag}-{traj}")
        get_prop = mk_get_prop(prop_cache, psel_pos_dict = psel_pos_dict)
        return get_prop
    return None

@q.timer
def run_job(job_tag, traj):
    q.check_stop()
    q.check_time_limit()
    #
    get_gf = run_gf(job_tag, traj)
    get_gt = run_gt(job_tag, traj, get_gf)
    get_eig = run_eig(job_tag, traj, get_gf)
    #
    get_prop = run_prop(job_tag, traj, get_gf, get_gt, get_eig)
    #
    if get_prop is not None:
        # auto_contractor_simple_test(job_tag, traj, get_prop)
        num_trials = 100
        # auto_contractor_vev(job_tag, traj, get_prop, num_trials)
        # auto_contractor_test_corr(job_tag, traj, get_prop, num_trials)
        # auto_contractor_meson_corr(job_tag, traj, get_prop, num_trials)
        # auto_contractor_pipi_corr(job_tag, traj, get_prop, num_trials)
        # auto_contractor_kpipi_corr(job_tag, traj, get_prop, num_trials)
        # auto_contractor_kpipi_corr_81oprs(job_tag, traj, get_prop, num_trials)
        auto_contractor_3f4f_matching(job_tag, traj, get_prop, num_trials)
    #
    q.clean_cache()

if __name__ == "__main__":
    qg.begin_with_gpt()
    job_tag = "test-4nt16"
    q.displayln_info(pprint.pformat(rup.dict_params[job_tag]))
    traj = 1000
    run_job(job_tag, traj)
    q.timer_display()
    qg.end_with_gpt()



#    Qlattice (https://github.com/jinluchang/qlattice)
#
#    Copyright (C) 2023
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from qlat_utils import \
        rel_mod, \
        rel_mod_sym, \
        c_rel_mod_sqr

auto_fac_funcs_list = [
        "rel_mod",
        "rel_mod_sym",
        "c_rel_mod_sqr",
        ]



from qlat_utils import *
import qlat.c as c

from qlat.geometry import *
from qlat.field import *
from qlat.field_selection import *
from qlat.utils_io import *

from qlat.c import SelectedField, SelectedFieldBase

@timer
def set_selected_field(sf, f):
    # deprecated use @=
    displayln_info("set_selected_field: deprecated")
    assert isinstance(sf, SelectedFieldBase)
    if isinstance(f, FieldBase):
        c.set_sfield_field(sf, f)
    elif isinstance(f, SelectedFieldBase):
        c.set_sfield_sfield(sf, f)
    else:
        raise Exception("set_selected_field")



### -------------------------------------------------------------------

cdef class FieldTYPENAME(FieldBase):

    ctype = ElemTypeTYPENAME

    def __cinit__(self):
        self.cdata = <long>&(self.xx)
        self.view_count = 0

    def __init__(self, Geometry geo = None, int multiplicity = 0):
        if geo is None:
            return
        if self.view_count > 0:
            raise ValueError("can't re-init while being viewed")
        self.xx.init(geo.xx, multiplicity)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef const cc.Geometry* p_geo = &self.xx.get_geo()
        cdef cc.Coordinate local_site = p_geo[0].local_site()
        cdef int multiplicity = p_geo[0].multiplicity
        cdef cc.Vector[cc.TYPENAME] fvec = cc.get_data(self.xx)
        cdef int ndim = 5 + ElemTypeTYPENAME.ndim()
        cdef char* fmt = ElemTypeTYPENAME.format()
        cdef Buffer buf = Buffer(self, ndim, ElemTypeTYPENAME.itemsize())
        cdef cc.std_vector[Py_ssize_t] vec = ElemTypeTYPENAME.shape()
        cdef Py_ssize_t* shape = &buf.shape_strides[0]
        cdef Py_ssize_t* strides = &buf.shape_strides[buf.ndim]
        cdef int i
        for i in range(4):
            shape[i] = local_site[i]
        shape[4] = multiplicity
        for i in range(ElemTypeTYPENAME.ndim()):
            shape[5 + i] = vec[i]
        buf.set_strides()
        buffer.buf = <char*>(fvec.data())
        if flags & PyBUF_FORMAT:
            buffer.format = fmt
        else:
            buffer.format = NULL
        buffer.internal = NULL
        buffer.itemsize = buf.itemsize
        buffer.len = buf.get_len()
        buffer.ndim = buf.ndim
        buffer.obj = buf
        buffer.readonly = 0
        buffer.shape = shape
        buffer.strides = strides
        buffer.suboffsets = NULL
        assert buffer.len * buffer.itemsize == fvec.size() * ElemTypeTYPENAME.size()
        self.view_count += 1

    def __releasebuffer__(self, Py_buffer *buffer):
        self.view_count -= 1

    def __imatmul__(self, f1):
        """
        f1 can be Field, SelectedField, SelectedPoints
        field geo does not change if already initialized
        """
        if isinstance(f1, FieldTYPENAME):
            self.xx = (<FieldTYPENAME>f1).xx
        elif isinstance(f1, SelectedFieldTYPENAME):
            c.set_field_sfield(self, f1)
        elif isinstance(f1, SelectedPointsTYPENAME):
            c.set_field_spfield(self, f1)
        else:
            raise Exception(f"FieldTYPENAME @= type mismatch {type(self)} {type(f1)}")
        return self

    @q.timer
    def copy(self, is_copying_data = True):
        f = type(self)()
        if is_copying_data:
            (<FieldTYPENAME>f).xx = self.xx
        return f

    @q.timer
    def set_zero(self):
        cc.set_zero(self.xx)

    def swap(self, FieldTYPENAME f1):
        cc.qswap(f1.xx, self.xx)

### -------------------------------------------------------------------

cdef class SelectedFieldTYPENAME(SelectedFieldBase):

    ctype = ElemTypeTYPENAME

    def __cinit__(self):
        self.cdata = <long>&(self.xx)
        self.view_count = 0

    def __init__(self, FieldSelection fsel, int multiplicity = 0):
        self.fsel = fsel
        if multiplicity > 0 and self.fsel is not None:
            if self.view_count > 0:
                raise ValueError("can't re-init while being viewed")
            self.xx.init(self.fsel.xx, multiplicity)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef long n_elems = self.xx.n_elems
        cdef int multiplicity = self.xx.get_geo().multiplicity
        cdef cc.Vector[cc.TYPENAME] fvec = cc.get_data(self.xx)
        cdef int ndim = 2 + ElemTypeTYPENAME.ndim()
        cdef char* fmt = ElemTypeTYPENAME.format()
        cdef Buffer buf = Buffer(self, ndim, ElemTypeTYPENAME.itemsize())
        cdef cc.std_vector[Py_ssize_t] vec = ElemTypeTYPENAME.shape()
        cdef Py_ssize_t* shape = &buf.shape_strides[0]
        cdef Py_ssize_t* strides = &buf.shape_strides[buf.ndim]
        cdef int i
        shape[0] = n_elems
        shape[1] = multiplicity
        for i in range(ElemTypeTYPENAME.ndim()):
            shape[2 + i] = vec[i]
        buf.set_strides()
        buffer.buf = <char*>(fvec.data())
        if flags & PyBUF_FORMAT:
            buffer.format = fmt
        else:
            buffer.format = NULL
        buffer.internal = NULL
        buffer.itemsize = buf.itemsize
        buffer.len = buf.get_len()
        buffer.ndim = buf.ndim
        buffer.obj = buf
        buffer.readonly = 0
        buffer.shape = shape
        buffer.strides = strides
        buffer.suboffsets = NULL
        assert buffer.len * buffer.itemsize == fvec.size() * ElemTypeTYPENAME.size()
        self.view_count += 1

    def __releasebuffer__(self, Py_buffer *buffer):
        self.view_count -= 1

    def __imatmul__(self, f1):
        """
        won't change self.fsel
        """
        if isinstance(f1, SelectedFieldTYPENAME):
            # two fsel do not need to match
            if self.fsel is f1.fsel:
                self.xx = (<SelectedFieldTYPENAME>f1).xx
            else:
                c.set_sfield_sfield(self, f1)
        elif isinstance(f1, FieldTYPENAME):
            c.set_sfield_field(self, f1)
        elif isinstance(f1, SelectedPointsTYPENAME):
            # psel may not have to be subset of fsel
            c.set_sfield_spfield(self, f1)
        else:
            raise Exception(f"SelectedFieldTYPENAME @= type mismatch {type(self)} {type(f1)}")
        return self

    @q.timer
    def copy(self, is_copying_data = True):
        f = type(self)(self.fsel)
        if is_copying_data:
            (<SelectedFieldTYPENAME>f).xx = self.xx
        return f

    @q.timer
    def set_zero(self):
        cc.set_zero(self.xx)

    def swap(self, SelectedFieldTYPENAME f1):
        assert f1.fsel is self.fsel
        cc.qswap(f1.xx, self.xx)

### -------------------------------------------------------------------

cdef class SelectedPointsTYPENAME(SelectedPointsBase):

    ctype = ElemTypeTYPENAME

    def __cinit__(self):
        self.cdata = <long>&(self.xx)
        self.view_count = 0

    def __init__(self, PointsSelection psel, int multiplicity = 0):
        self.psel = psel
        if multiplicity > 0 and self.psel is not None:
            if self.view_count > 0:
                raise ValueError("can't re-init while being viewed")
            self.xx.init(self.psel.xx, multiplicity)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef long n_points = self.xx.n_points
        cdef int multiplicity = self.xx.multiplicity
        cdef cc.Vector[cc.TYPENAME] fvec = cc.get_data(self.xx)
        cdef int ndim = 2 + ElemTypeTYPENAME.ndim()
        cdef char* fmt = ElemTypeTYPENAME.format()
        cdef Buffer buf = Buffer(self, ndim, ElemTypeTYPENAME.itemsize())
        cdef cc.std_vector[Py_ssize_t] vec = ElemTypeTYPENAME.shape()
        cdef Py_ssize_t* shape = &buf.shape_strides[0]
        cdef Py_ssize_t* strides = &buf.shape_strides[buf.ndim]
        cdef int i
        shape[0] = n_points
        shape[1] = multiplicity
        for i in range(ElemTypeTYPENAME.ndim()):
            shape[2 + i] = vec[i]
        buf.set_strides()
        buffer.buf = <char*>(fvec.data())
        if flags & PyBUF_FORMAT:
            buffer.format = fmt
        else:
            buffer.format = NULL
        buffer.internal = NULL
        buffer.itemsize = buf.itemsize
        buffer.len = buf.get_len()
        buffer.ndim = buf.ndim
        buffer.obj = buf
        buffer.readonly = 0
        buffer.shape = shape
        buffer.strides = strides
        buffer.suboffsets = NULL
        assert buffer.len * buffer.itemsize == fvec.size() * ElemTypeTYPENAME.size()
        self.view_count += 1

    def __releasebuffer__(self, Py_buffer *buffer):
        self.view_count -= 1

    def __imatmul__(self, f1):
        """
        won't change self.psel
        """
        if isinstance(f1, SelectedPointsTYPENAME):
            # two psel must be the same object
            if self.psel is f1.psel:
                self.xx = (<SelectedPointsTYPENAME>f1).xx
            else:
                raise Exception("SelectedPointsTYPENAME @= psel mismatch")
        elif isinstance(f1, SelectedFieldTYPENAME):
            # only assign available points
            c.set_spfield_sfield(self, f1)
        elif isinstance(f1, FieldTYPENAME):
            c.set_spfield_field(self, f1)
        else:
            raise Exception(f"SelectedPointsTYPENAME @= type mismatch {type(self)} {type(f1)}")
        return self

    @q.timer
    def copy(self, is_copying_data = True):
        f = type(self)(self.psel)
        if is_copying_data:
            (<SelectedPointsTYPENAME>f).xx = self.xx
        return f

    @q.timer
    def set_zero(self):
        cc.set_zero(self.xx)

    def swap(self, SelectedPointsTYPENAME f1):
        assert f1.psel is self.psel
        cc.qswap(f1.xx, self.xx)

### -------------------------------------------------------------------

field_type_dict[ElemTypeTYPENAME] = FieldTYPENAME

selected_field_type_dict[ElemTypeTYPENAME] = SelectedFieldTYPENAME

selected_points_type_dict[ElemTypeTYPENAME] = SelectedPointsTYPENAME

### -------------------------------------------------------------------



import qlat.c as c

from qlat.qcd import *
from qlat.propagator import *

def gf_ape_smear(gf : GaugeField, alpha : float, steps : int = 1):
    return c.gf_ape_smear(gf, gf, alpha, steps)

def gf_spatial_ape_smear(gf : GaugeField, alpha : float, steps : int = 1):
    # used value: alpha = 0.5, steps = 30
    return c.gf_spatial_ape_smear(gf, gf, alpha, steps)

def gf_hyp_smear(gf : GaugeField, alpha1 : float, alpha2 : float, alpha3 : float):
    # values in paper is 0.75 0.6 0.3
    # 10.1103/PhysRevD.64.034504 Eq(4)
    return c.gf_hyp_smear(gf, gf, alpha1, alpha2, alpha3)

def prop_smear(prop : Prop, gf1 : GaugeField, coef : float, step : int, mom = None, smear_in_time_dir : bool = False, mode_smear : int = 1):
    # gf_ape = gf_spatial_ape_smear(gf, 0.5, 30)
    # gf1 = mk_left_expanded_gauge_field(gf_ape)
    # mom: momentum smearing in lattice unit 1/a (not unit of lattice momentum 2 pi / L / a)
    # smear params:
    # 24D: coef = 0.9375, step = 10
    # 48I: coef = 0.9375, step = 29
    # mom = 0.5 * mom of the corresponding hadron
    if mom is None:
        mom = [ 0.0, 0.0, 0.0, 0.0, ]
    return c.prop_smear(prop, gf1, coef, step, mom, smear_in_time_dir, mode_smear)



# Author: Christoph Lehner 2021
# Author: Luchang Jin 2023

import qlat as q
import sys

if len(sys.argv) == 1:
    q.displayln_info("Usage: gauge-fix-coulomb [ --source source_config ] [ --output gt.field ] [ --guess gt-guess.field ]")
    q.displayln_info("If the output file already exist, it will only check output file.")
    q.displayln_info("Default output file is 'gt.field'.")
    q.displayln_info("Will use CPS gauge fix format is output file end with '.gfix'.")
    sys.exit()

import gpt as g
import qlat_gpt as qg

qg.begin_with_gpt()

# Parameters
p_mpi_split = g.default.get_ivec("--mpi_split", [ 1, 1, 1, ], 3)
p_maxiter_cg = g.default.get_int("--maxiter_cg", 200)
p_maxiter_gd = g.default.get_int("--maxiter_gd", 10)
p_maxcycle_cg = g.default.get_int("--maxcycle_cg", 50)
p_log_every = g.default.get_int("--log_every", 1)
p_eps = g.default.get_float("--eps", 1e-12)
p_step = g.default.get_float("--step", 0.3)
p_step_gd = g.default.get_float("--step_gd", 0.1)
p_source = g.default.get("--source", None)
p_guess = g.default.get("--guess", None)
p_output = g.default.get("--output", "gt.field")
p_rng_seed = g.default.get("--random", None)

def load_gf():
    if p_source is None:
        q.displayln_info("Need to provide source file with '--source filename'. Use a sample gauge field for now.")
        total_site = [ 4, 4, 4, 8, ]
        geo = q.Geometry(total_site, 1)
        gf = q.GaugeField(geo)
        rs = q.RngState("seed")
        gf.set_rand(rs.split("gf-init"), 0.5, 10)
    else:
        q.displayln_info(f"Loading gauge field '{p_source}'.")
        if not q.does_file_exist_qar_sync_node(p_source):
            q.displayln_info(f"ERROR: Gauge field file '{p_source}' does not exist.")
            sys.exit(1)
        gf = q.GaugeField()
        gf.load(p_source)
    return gf

def load_gt_guess():
    if p_guess is None:
        q.displayln_info("No guess is provided. Gauge fix from scratch.")
        return None
    elif not q.does_file_exist_qar_sync_node(p_guess):
        q.displayln_info(f"WARNING: Guess file '{p_guess}' does not exist. Gauge fix from scratch.")
        return None
    q.displayln_info(f"Find '{p_guess}'.")
    gt = q.GaugeTransform()
    if p_guess.endswith(".gfix"):
        q.displayln_info(f"Assuming CPS gauge fix format.")
        gt.load_cps(p_guess)
    else:
        q.displayln_info(f"Assuming Qlat field format (double precision, big endian).")
        gt.load_double(p_guess)
    return gt

def load_gt_output():
    assert isinstance(p_output, str)
    if not q.does_file_exist_qar_sync_node(p_output):
        q.displayln_info(f"Output file '{p_output}' does not exist. Will do gauge fix.")
        return None
    q.displayln_info(f"Find '{p_output}'. Will only do checking.")
    gt = q.GaugeTransform()
    if p_output.endswith(".gfix"):
        q.displayln_info(f"Assuming CPS gauge fix format.")
        gt.load_cps(p_output)
    else:
        q.displayln_info(f"Assuming Qlat field format (double precision, big endian).")
        gt.load_double(p_output)
    return gt

def save_gt_output(gt):
    assert isinstance(p_output, str)
    assert not q.does_file_exist_sync_node(p_output)
    q.displayln_info(f"Start to write to '{p_output}'.")
    if p_output.endswith(".gfix"):
        q.displayln_info(f"Assuming CPS gauge fix format.")
        gt.save_cps(p_output)
    else:
        q.displayln_info(f"Assuming Qlat field format (double precision, big endian).")
        gt.save_double(p_output)
    q.sync_node()

gf = load_gf()

gt = load_gt_output()

if gt is None:
    # Do gauge fixing
    gt = load_gt_guess()
    gt = qg.gauge_fix_coulomb(
            gf,
            gt=gt,
            mpi_split=p_mpi_split,
            maxiter_gd=p_maxiter_gd,
            maxiter_cg=p_maxiter_cg,
            maxcycle_cg=p_maxcycle_cg,
            log_every=p_log_every,
            eps=p_eps,
            step=p_step,
            step_gd=p_step_gd,
            rng_seed=p_rng_seed,
            )
    save_gt_output(gt)
else:
    # Only do checking
    qg.check_gauge_fix_coulomb(gf, gt, eps=p_eps)

q.timer_display()

qg.end_with_gpt()

exit()



from .jobs import *

# ----

@q.timer
def get_all_points(total_site):
    n_points = total_site[0] * total_site[1] * total_site[2] * total_site[3]
    xg_list = []
    for index in range(n_points):
        xg = q.Coordinate()
        xg.from_index(index, total_site)
        xg_list.append(xg)
    return xg_list

@q.timer
def get_all_points_psel(total_site):
    geo = q.Geometry(total_site, 1)
    xg_list = get_all_points(total_site)
    psel = q.PointsSelection([ xg.to_list() for xg in xg_list ], geo)
    return psel

# ----

@q.timer
def run_get_inverter_checker(job_tag, traj, *, inv_type, get_gf, get_gt = None, get_eig = None):
    if None in [ get_gf, ]:
        return
    if get_gt is None:
        get_gt = lambda: None
    if get_eig is None:
        get_eig = lambda: None
    gf = get_gf()
    gt = get_gt()
    eig = get_eig()
    inv_acc = 2
    from . import rbc_ukqcd as ru
    ru.get_inv(gf, job_tag, inv_type, inv_acc, gt = gt, eig = eig)

@q.timer_verbose
def compute_prop_1_checker(inv, src, *, tag, sfw, path_sp):
    fn_spw = os.path.join(path_sp, f"{tag} ; wsnk.lat")
    sol = inv * src
    sol.save_float_from_double(sfw, tag)
    sfw.flush()
    sol_ws = sol.glb_sum_tslice()
    sol_ws.save(get_save_path(fn_spw))
    return sol

@q.timer
def compute_prop_wsrc_checker(job_tag, tslice, inv_type, inv_acc, *,
                              idx, gf, gt, sfw, path_sp, eig, finished_tags):
    from . import rbc_ukqcd as ru
    tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in finished_tags:
        return None
    q.displayln_info(f"compute_prop_wsrc: idx={idx} tslice={tslice}", job_tag, inv_type, inv_acc)
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, gt = gt, eig = eig)
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site, 1)
    src = q.mk_wall_src(geo, tslice)
    prop = compute_prop_1_checker(inv, src, tag = tag, sfw = sfw, path_sp = path_sp)

@q.timer_verbose
def compute_prop_wsrc_all_checker(job_tag, traj, *,
                                  inv_type, gf, gt, eig):
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    path_s = f"{job_tag}/prop-wsrc-{inv_type_name}/traj-{traj}"
    path_sp = f"{job_tag}/psel-prop-wsrc-{inv_type_name}/traj-{traj}"
    finished_tags = q.properly_truncate_fields(get_save_path(path_s + ".acc"))
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", [ 1, 1, 1, 4, ])
    inv_acc = 2
    for idx, tslice in enumerate(range(total_site[3])):
        compute_prop_wsrc_checker(job_tag, tslice, inv_type, inv_acc = 2,
                                  idx = idx, gf = gf, gt = gt, sfw = sfw, path_sp = path_sp,
                                  eig = eig, finished_tags = finished_tags)
    sfw.close()
    q.qtouch_info(get_save_path(os.path.join(path_sp, "checkpoint.txt")))
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
    q.qar_create_info(get_save_path(path_sp + ".qar"), get_save_path(path_sp), is_remove_folder_after = True)
    q.qar_create_info(get_save_path(path_s + ".qar"), get_save_path(path_s), is_remove_folder_after = True)

@q.timer
def run_prop_wsrc_checker(job_tag, traj, *, inv_type, get_gf, get_eig, get_gt):
    if None in [ get_gf, get_gt, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    if get_load_path(f"{job_tag}/prop-wsrc-{inv_type_name}/traj-{traj}/geon-info.txt") is not None:
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-wsrc-{inv_type_name}"):
        gf = get_gf()
        gt = get_gt()
        eig = get_eig()
        compute_prop_wsrc_all_checker(job_tag, traj,
                                      inv_type = inv_type, gf = gf, gt = gt, eig = eig)
        q.release_lock()

# ----

@q.timer_verbose
def compute_prop_2_checker(inv, src, *, tag, sfw):
    sol = inv * src
    sol.save_float_from_double(sfw, tag)
    sfw.flush()
    return sol

@q.timer
def compute_prop_psrc_checker(job_tag, xg_src, inv_type, inv_acc, *,
                              idx, gf, gt, sfw, eig, finished_tags):
    from . import rbc_ukqcd as ru
    xg = xg_src.to_list()
    xg_str = f"({xg[0]},{xg[1]},{xg[2]},{xg[3]})"
    tag = f"xg={xg_str} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in finished_tags:
        return None
    q.displayln_info(f"compute_prop_psrc: {job_tag} idx={idx} tag='{tag}'")
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, gt = gt, eig = eig)
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    geo = q.Geometry(total_site, 1)
    src = q.mk_point_src(geo, xg)
    prop = compute_prop_2_checker(inv, src, tag = tag, sfw = sfw)

@q.timer_verbose
def compute_prop_psrc_all_checker(job_tag, traj, *,
                                  inv_type, gf, gt, eig):
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    path_s = f"{job_tag}/prop-psrc-{inv_type_name}/traj-{traj}"
    finished_tags = q.properly_truncate_fields(get_save_path(path_s + ".acc"))
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", [ 1, 1, 1, 4, ])
    def comp(idx, xg_src, inv_acc):
        compute_prop_psrc_checker(job_tag, xg_src, inv_type, inv_acc,
                idx = idx, gf = gf, gt = gt, sfw = sfw,
                eig = eig, finished_tags = finished_tags)
    for idx, xg_src in enumerate(get_all_points(total_site)):
        comp(idx, xg_src, inv_acc = 2)
    sfw.close()
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
    q.qar_create_info(get_save_path(path_s + ".qar"), get_save_path(path_s), is_remove_folder_after = True)

@q.timer
def run_prop_psrc_checker(job_tag, traj, *, inv_type, get_gf, get_eig, get_gt):
    if None in [ get_gf, get_gt, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    if get_load_path(f"{job_tag}/prop-psrc-{inv_type_name}/traj-{traj}/geon-info.txt") is not None:
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-psrc-{inv_type_name}"):
        gf = get_gf()
        gt = get_gt()
        eig = get_eig()
        compute_prop_psrc_all_checker(job_tag, traj,
                                      inv_type = inv_type, gf = gf, gt = gt,
                                      eig = eig)
        q.release_lock()

# ----

@q.timer_verbose
def load_prop_psrc(job_tag, traj, inv_type):
    inv_tag_list = [ "l", "s", ]
    inv_tag = inv_tag_list[inv_type]
    cache = q.mk_cache(f"prop_cache", f"{job_tag}", f"{traj}", inv_tag)
    if "psnk-psrc" in cache:
        return
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    inv_acc = 2
    path_s = f"{job_tag}/prop-psrc-{inv_type_name}/traj-{traj}/geon-info.txt"
    psel = get_all_points_psel(total_site)
    prop_list = []
    xg_list = [ q.Coordinate(xg) for xg in psel.to_list() ]
    sfr = q.open_fields(get_load_path(path_s), "r")
    for xg_src in xg_list:
        xg_idx = xg_src.to_index(total_site)
        xg = xg_src.to_list()
        xg_str = f"({xg[0]},{xg[1]},{xg[2]},{xg[3]})"
        tag = f"xg={xg_str} ; type={inv_type} ; accuracy={inv_acc}"
        prop = q.Prop()
        prop.load_double_from_float(sfr, tag)
        sp_prop = q.PselProp(psel)
        sp_prop @= prop
        prop_list.append(sp_prop)
    sfr.close()
    cache["psnk-psrc"] = prop_list

@q.timer_verbose
def load_prop_wsrc(job_tag, traj, inv_type):
    inv_tag_list = [ "l", "s", ]
    inv_tag = inv_tag_list[inv_type]
    cache = q.mk_cache(f"prop_cache", f"{job_tag}", f"{traj}", inv_tag)
    if "psnk-wsrc" in cache and "wsnk-wsrc" in cache:
        return
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    inv_acc = 2
    path_s = f"{job_tag}/prop-wsrc-{inv_type_name}/traj-{traj}/geon-info.txt"
    path_sp = f"{job_tag}/psel-prop-wsrc-{inv_type_name}/traj-{traj}/"
    psel = get_all_points_psel(total_site)
    psel_ts = q.get_psel_tslice(total_site)
    prop_list = []
    prop2_list = []
    tslice_list = list(range(total_site[3]))
    sfr = q.open_fields(get_load_path(path_s), "r")
    for tslice in tslice_list:
        tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
        prop = q.Prop()
        prop.load_double_from_float(sfr, tag)
        sp_prop = q.PselProp(psel)
        sp_prop @= prop
        prop_list.append(sp_prop)
        fn_spw = os.path.join(path_sp, f"{tag} ; wsnk.lat")
        sp_prop2 = q.PselProp(psel_ts)
        sp_prop2.load(get_load_path(fn_spw))
        prop2_list.append(sp_prop2)
    sfr.close()
    cache["psnk-wsrc"] = prop_list
    cache["wsnk-wsrc"] = prop2_list

@q.timer_verbose
def run_get_prop_checker(job_tag, traj, *,
                         get_gf,
                         get_gt):
    traj_gf = traj
    fns_props = [
            (f"{job_tag}/prop-psrc-light/traj-{traj_gf}.qar", f"{job_tag}/prop-psrc-light/traj-{traj_gf}/geon-info.txt",),
            (f"{job_tag}/prop-psrc-strange/traj-{traj_gf}.qar", f"{job_tag}/prop-psrc-strange/traj-{traj_gf}/geon-info.txt",),
            (f"{job_tag}/prop-wsrc-light/traj-{traj_gf}.qar", f"{job_tag}/prop-wsrc-light/traj-{traj_gf}/geon-info.txt",),
            (f"{job_tag}/prop-wsrc-strange/traj-{traj_gf}.qar", f"{job_tag}/prop-wsrc-strange/traj-{traj_gf}/geon-info.txt",),
            (f"{job_tag}/psel-prop-wsrc-light/traj-{traj_gf}.qar", f"{job_tag}/psel-prop-wsrc-light/traj-{traj_gf}/checkpoint.txt",),
            (f"{job_tag}/psel-prop-wsrc-strange/traj-{traj_gf}.qar", f"{job_tag}/psel-prop-wsrc-strange/traj-{traj_gf}/checkpoint.txt",),
            ]
    for fn in fns_props:
        if get_load_path(fn) is None:
            return None
    @q.timer_verbose
    def mk_get_prop():
        q.timer_fork()
        total_site = q.Coordinate(get_param(job_tag, "total_site"))
        gf = get_gf()
        gt = get_gt()
        #
        load_prop_psrc(job_tag, traj, inv_type = 0)
        load_prop_psrc(job_tag, traj, inv_type = 1)
        load_prop_wsrc(job_tag, traj, inv_type = 0)
        load_prop_wsrc(job_tag, traj, inv_type = 1)
        #
        prop_cache = q.mk_cache(f"prop_cache", f"{job_tag}", f"{traj}")
        def get_prop(flavor, p_snk, p_src):
            cache = prop_cache[flavor]
            p_snk_tag, p_snk_xg = p_snk
            p_src_tag, p_src_xg = p_src
            if p_snk_tag == "point" and p_src_tag == "point":
                prop_list = cache["psnk-psrc"]
                p_src_idx = p_src_xg.to_index(total_site)
                p_snk_idx = p_snk_xg.to_index(total_site)
                return prop_list[p_src_idx].get_elem_wm(p_snk_idx)
            elif p_snk_tag == "point" and p_src_tag == "wall":
                prop_list = cache["psnk-wsrc"]
                p_snk_idx = p_snk_xg.to_index(total_site)
                return prop_list[p_src_xg].get_elem_wm(p_snk_idx)
            elif p_snk_tag == "wall" and p_src_tag == "point":
                prop_list = cache["psnk-wsrc"]
                p_src_idx = p_src_xg.to_index(total_site)
                return q.wilson_matrix_g5_herm(prop_list[p_snk_xg].get_elem_wm(p_src_idx))
            elif p_snk_tag == "wall" and p_src_tag == "wall":
                prop_list = cache["wsnk-wsrc"]
                return prop_list[p_src_xg].get_elem_wm(p_snk_xg)
            else:
                raise Exception(f"get_prop: f={flavor} snk={p_snk} src={p_src}")
        #
        q.timer_display()
        q.timer_merge()
        return get_prop
    return q.lazy_call(mk_get_prop)

# ----



import qlat.c as c

from qlat.qcd import *
from qlat.hmc import *

def gf_energy_density(gf : GaugeField):
    return c.gf_energy_density(gf)

def gf_wilson_flow_force(gf, c1 = 0.0):
    ga = GaugeAction(3.0, c1)
    gm_force = GaugeMomentum()
    set_gm_force(gm_force, gf, ga)
    return gm_force

def gf_wilson_flow_step(gf, epsilon, *, c1 = 0.0, wilson_flow_integrator_type = "runge-kutta"):
    if wilson_flow_integrator_type == "runge-kutta":
        return gf_wilson_flow_runge_kutta(gf, epsilon, c1 = c1)
    elif wilson_flow_integrator_type == "euler":
        return gf_wilson_flow_euler(gf, epsilon, c1 = c1)

def gf_wilson_flow_runge_kutta(gf, epsilon, *, c1 = 0.0):
    # Runge-Kutta scheme
    # http://arxiv.org/abs/1006.4518v3
    return c.gf_wilson_flow_step(gf, epsilon, c1)

def gf_wilson_flow_euler(gf, epsilon, *, c1 = 0.0):
    force = gf_wilson_flow_force(gf, c1 = c1)
    return gf_evolve(gf, force, epsilon)

@timer
def gf_wilson_flow(gf: GaugeField, flow_time : float, steps : int,
        *, c1 : float = 0.0, existing_flow_time : float = 0.0, wilson_flow_integrator_type = "runge-kutta"):
    epsilon = flow_time / steps
    energy_density_list = []
    for i in range(steps):
        gf_wilson_flow_step(gf, epsilon, c1 = c1, wilson_flow_integrator_type = wilson_flow_integrator_type)
        t = (i + 1) * epsilon + existing_flow_time
        energy_density = gf_energy_density(gf)
        energy_density_list.append(energy_density)
        displayln_info(f"gf_wilson_flow: t={t} ; E={energy_density} ; t^2 E={t*t*energy_density}")
    return energy_density_list



from .. import rbc_ukqcd_params as rup

job_tag = "32IH2"

dict_params = {}

rup.dict_params[job_tag] = dict_params

dict_params["job_tag"] = job_tag

dict_params["total_site"] = [ 32, 32, 32, 64, ]

dict_params["load_config_params"] = { "twist_boundary_at_boundary":[ 0.0, 0.0, 0.0, -0.5, ] }

def mk_fermion_params(inv_type, inv_acc):
    params = {}
    params["M5"] = 1.8
    params["boundary_phases"] = [1.0, 1.0, 1.0, 1.0] # twist boundary after loading configuration
    params["b"] = 1.0
    params["c"] = 0.0
    params["Ls"] = 16
    if inv_type == 0:
        params["mass"] = 0.006
    elif inv_type == 1:
        params["mass"] = 0.03
    elif inv_type == 2:
        params["mass"] = 0.31
    else:
        assert False
    return params

def mk_dict_fermion_params():
    params = {}
    for inv_type in [ 0, 1, 2, ]:
        params[inv_type] = {}
        for inv_acc in [ 0, 1, 2, ]:
            params[inv_type][inv_acc] = mk_fermion_params(inv_type, inv_acc)
    return params

dict_params["fermion_params"] = mk_dict_fermion_params()

dict_params[f"cg_params-1-0"] = {}
dict_params[f"cg_params-1-0"]["maxiter"] = 250

dict_params[f"cg_params-1-1"] = {}
dict_params[f"cg_params-1-1"]["maxiter"] = 250

def mk_lanc_params(inv_type, inv_acc):
    assert inv_acc == 0
    if inv_type == 0:
        c_low = 0.0009
        n_stop = 250
        n_keep = n_stop + 10
        n_max = n_stop + 50
    elif inv_type == 1:
        c_low = 0.0007
        n_stop = 150
        n_keep = n_stop + 10
        n_max = n_stop + 50
    else:
        assert False
    fermion_params = mk_dict_fermion_params()[inv_type][inv_acc]
    pit_params = { "eps": 0.01, "maxiter": 500, "real": True }
    cheby_params = {"low": c_low, "high": 5.5, "order": 200}
    irl_params = {
            "Nstop": n_stop,
            "Nk": n_keep,
            "Nm": n_max,
            "resid": 1e-8,
            "betastp": 0.0,
            "maxiter": 20,
            "Nminres": 0,
            # "maxapply": 100
            }
    return {
            "fermion_params": fermion_params,
            "pit_params": pit_params,
            "cheby_params": cheby_params,
            "irl_params": irl_params,
            }

def mk_clanc_params(inv_type, inv_acc):
    assert inv_acc == 0
    if inv_type == 0:
        nbasis = 250
        c_low = 0.002
        n_stop = 500
        n_keep = n_stop + 20
        n_max = n_stop + 150
        n_single = 100
    elif inv_type == 1:
        nbasis = 150
        c_low = 0.0014
        n_stop = 300
        n_keep = n_stop + 20
        n_max = n_stop + 100
        n_single = 50
    else:
        assert False
    block = [ 4, 4, 4, 4, ]
    cheby_params = { "low": c_low, "high": 5.5, "order": 100, }
    irl_params = {
            "Nstop": n_stop,
            "Nk": n_keep,
            "Nm": n_max,
            "resid": 1e-8,
            "betastp": 0.0,
            "maxiter": 20,
            "Nminres": 0,
            #    "maxapply" : 100
            }
    smoother_params = { "eps": 1e-6, "maxiter": 25, }
    save_params = { "nsingle": n_single, "mpi": [ 1, 1, 1, 16, ], }
    return {
            "block": block,
            "nbasis": nbasis,
            "cheby_params": cheby_params,
            "irl_params": irl_params,
            "smoother_params": smoother_params,
            "save_params": save_params,
            }

dict_params["lanc_params"] = { 0: { 0: mk_lanc_params(0, 0), }, 1: { 0: mk_lanc_params(1, 0), }, }

dict_params["clanc_params"] = { 0: { 0: mk_clanc_params(0, 0), }, 1: { 0: mk_clanc_params(1, 0), }, }



from qlat_utils import *
import qlat.c as c

from qlat.propagator import *

@timer
def contract_chvp3_field(prop1, prop2, tslice):
    ld = LatData()
    c.contract_chvp3_sfield(ld, prop1, prop2, tslice)
    return ld



import qlat.c as c

class GaugeAction:

    def __init__(self, beta, c1 = 0.0):
        self.cdata = c.mk_gauge_action(beta, c1)

    def __del__(self):
        assert isinstance(self.cdata, int)
        c.free_gauge_action(self)

    def __imatmul__(self, v1):
        assert isinstance(v1, GaugeAction)
        c.set_gauge_action(self, v1)
        return self

    def beta(self):
        return c.get_beta_gauge_action(self)

    def c1(self):
        return c.get_c1_gauge_action(self)




from qlat_utils import *
import qlat.c as c

import os
import pickle

from qlat.c import check_time_limit, check_stop
from qlat.c import get_time_limit, get_default_budget


from qlat.mpi import *

def obtain_lock(path):
    mk_file_dirs_info(path)
    return c.obtain_lock(path)

def release_lock():
    return c.release_lock()

def qquit(msg):
    """
    clean python cache and then call c.qquit(msg) (which clear all the C++ level cache and then quit)
    """
    clean_cache()
    return c.qquit(msg)

def get_remaining_time():
    return get_time_limit() - get_actual_total_time()



import qlat.c as c

from qlat.qcd import *
from qlat.wilson_flow import *

from pprint import pformat

def gf_topology_field_clf(gf : GaugeField):
    geo = gf.geo()
    topf = FieldDouble(geo, 1)
    c.gf_topology_field_clf(topf, gf)
    return topf

@timer
def gf_topology_clf(gf : GaugeField):
    return gf_topology_field_clf(gf).glb_sum()[0]

def gf_topology_field(gf : GaugeField):
    geo = gf.geo()
    topf = FieldDouble(geo, 1)
    c.gf_topology_field(topf, gf)
    return topf

@timer
def gf_topology(gf : GaugeField):
    return gf_topology_field(gf).glb_sum()[0]

def gf_topology_terms_field(gf : GaugeField):
    geo = gf.geo()
    topf = FieldDouble(geo, 5)
    c.gf_topology_terms_field(topf, gf)
    return topf

@timer
def gf_topology_terms(gf : GaugeField):
    return gf_topology_terms_field(gf).glb_sum()

@timer_verbose
def smear_measure_topo(gf, smear_info_list = None, *, is_show_topo_terms = False):
    # smear_info = [ [ step_size, n_step, c1 = 0.0, wilson_flow_integrator_type = "runge-kutta", ], ... ]
    # c1 = 0.0 # Wilson
    # c1 = -0.331 # Iwasaki
    # c1 = -1.4008 # DBW2
    # wilson_flow_integrator_type = "runge-kutta"
    # wilson_flow_integrator_type = "euler"
    if smear_info_list is None:
        smear_info_list = [
                [ 0.05, 20, 0.0, "euler", ],
                [ 0.05, 20, 0.0, "euler", ],
                [ 0.05, 20, 0.0, "euler", ],
                [ 0.01, 50, -1.4008, "euler", ],
                [ 0.01, 50, -1.4008, "euler", ],
                [ 0.01, 50, -1.4008, "euler", ],
                [ 0.01, 50, -1.4008, "euler", ],
                ]
    displayln_info(f"smear_info_list =")
    displayln_info(pformat(smear_info_list))
    flow_time = 0
    topo_list = []
    @timer
    def smear(step_size, n_step, c1 = 0.0, wilson_flow_integrator_type = "runge-kutta"):
        nonlocal flow_time
        flow_time += n_step * step_size
        for i in range(n_step):
            gf_wilson_flow_step(gf, step_size, c1 = c1, wilson_flow_integrator_type = wilson_flow_integrator_type)
    @timer
    def measure():
        gf.show_info()
        plaq = gf.plaq()
        energy_density = gf_energy_density(gf)
        topo_field_clf = gf_topology_field_clf(gf)
        topo_clf = topo_field_clf.glb_sum().item()
        t_sum_clf = topo_field_clf.glb_sum_tslice()
        t_sum_clf = [ t_sum_clf.get_elem(t).item() for t in range(t_sum_clf.n_points()) ]
        topo_field = gf_topology_field(gf)
        topo = topo_field.glb_sum().item()
        t_sum = topo_field.glb_sum_tslice()
        t_sum = [ t_sum.get_elem(t).item() for t in range(t_sum.n_points()) ]
        displayln_info(f"t={flow_time} topo_clf={topo_clf} topo={topo}")
        displayln_info(pformat(list(enumerate(zip(t_sum_clf, t_sum)))))
        topo_list.append({
            "flow_time": flow_time,
            "plaq": plaq,
            "energy_density": energy_density,
            "topo_clf": topo_clf,
            "topo_clf_tslice": t_sum_clf,
            "topo": topo,
            "topo_tslice": t_sum,
            })
        if is_show_topo_terms:
            topo_terms = gf_topology_terms(gf)
            displayln_info(f"t={t} topo={topo} {sum(topo_terms)}")
            topo_terms_str = ',\n '.join([ str(x) for x in topo_terms ])
            displayln_info(f"[ {topo_terms_str},\n]")
    measure()
    for si in smear_info_list:
        smear(*si)
        measure()
    return topo_list



from qlat_utils import *
import qlat.c as c

from qlat.field import *
from qlat.field_selection import *
from qlat.selected_field import *
from qlat.utils_io import *

cache_fields_io = mk_cache("fields_io")

class ShuffledFieldsWriter:

    # self.cdata

    def __init__(self, path, new_size_node, is_append = False):
        assert isinstance(path, str)
        assert isinstance(is_append , bool)
        self.cdata = c.mk_sfw(path, new_size_node, is_append)

    def close(self):
        if self.cdata is not None:
            c.free_sfw(self)
        self.cdata = None
        cache_fields_io.pop(id(self), None)

    def __del__(self):
        self.close()

    def new_size_node(self):
        return c.get_new_size_node_sfw(self)

    def get_cache_sbs(self, fsel):
        if id(self) in cache_fields_io:
            [ c_fsel, c_sbs ] = cache_fields_io[id(self)]
            if fsel is c_fsel:
                return c_sbs
        sbs = ShuffledBitSet(fsel, self.new_size_node())
        cache_fields_io[id(self)] = [ fsel, sbs, ]
        return sbs

    def write(self, fn, obj):
        assert isinstance(fn, str)
        if isinstance(obj, FieldBase):
            return c.write_sfw_field(self, fn, obj)
        elif isinstance(obj, SelectedFieldBase):
            return c.write_sfw_sfield(self, fn, obj, self.get_cache_sbs(obj.fsel))
        else:
            raise Exception("ShuffledFieldsWriter.save")

    def flush(self):
        return c.flush_sfw(self)

class ShuffledFieldsReader:

    # self.cdata
    # self.tags

    def __init__(self, path, new_size_node = None):
        assert isinstance(path, str)
        if new_size_node is None:
            self.cdata = c.mk_sfr(path)
        else:
            self.cdata = c.mk_sfr(path, new_size_node)
        self.tags = None

    def close(self):
        if self.cdata is not None:
            c.free_sfr(self)
        self.cdata = None
        self.tags = None
        cache_fields_io.pop(id(self), None)

    def __del__(self):
        self.close()

    def new_size_node(self):
        return c.get_new_size_node_sfr(self)

    def get_cache_sbs(self, fsel):
        if id(self) in cache_fields_io:
            [ c_fsel, c_sbs ] = cache_fields_io[id(self)]
            if fsel is c_fsel:
                return c_sbs
        sbs = ShuffledBitSet(fsel, self.new_size_node())
        cache_fields_io[id(self)] = [ fsel, sbs, ]
        return sbs

    def read(self, fn, obj):
        # Can also read SelectedField obj with obj.fsel is None
        # After reading, obj.fsel will be properly loaded.
        assert isinstance(fn, str)
        if isinstance(obj, FieldBase):
            return c.read_sfr_field(self, fn, obj)
        elif isinstance(obj, SelectedFieldBase):
            fsel = obj.fsel
            if fsel is None:
                if obj.view_count > 0:
                    raise ValueError("can't re-init while being viewed")
                obj.fsel = FieldSelection()
                return c.read_sfr_sfield(self, fn, None, obj, obj.fsel)
            else:
                return c.read_sfr_sfield(self, fn, self.get_cache_sbs(obj.fsel), obj)
        else:
            raise Exception("ShuffledFieldsReader.load")

    def list(self):
        return c.list_sfr(self)

    def has_sync_node(self, fn):
        return c.does_file_exist_sync_node_sfr(self, fn)

    def has(self, fn):
        if self.tags is None:
            self.tags = set(self.list())
        return fn in self.tags

class ShuffledBitSet:

    def __init__(self, fsel, new_size_node):
        assert isinstance(fsel, FieldSelection)
        self.cdata = c.mk_sbs(fsel, new_size_node)

    def __del__(self):
        assert isinstance(self.cdata, int)
        c.free_sbs(self)

def open_fields(path, mode, new_size_node = None):
    # path can be the folder path or the 'geon-info.txt' path
    assert isinstance(path, str)
    assert isinstance(mode, str)
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    if mode == "r":
        return ShuffledFieldsReader(path, new_size_node)
    elif mode == "w":
        assert new_size_node is not None
        return ShuffledFieldsWriter(path, new_size_node)
    elif mode == "a":
        if new_size_node is None:
            return ShuffledFieldsWriter(path, [0, 0, 0, 0], True)
        else:
            return ShuffledFieldsWriter(path, new_size_node, True)
    else:
        raise Exception("open_fields")

def list_fields(path, new_size_node = None):
    assert isinstance(path, str)
    sfr = open_fields(path, "r", new_size_node)
    fns = sfr.list()
    sfr.close()
    return fns

def properly_truncate_fields(path, is_check_all = False, is_only_check = False, new_size_node = None):
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    if new_size_node is None:
        return c.properly_truncate_fields_sync_node(path, is_check_all, is_only_check)
    else:
        return c.properly_truncate_fields_sync_node(path, is_check_all, is_only_check, new_size_node)

def truncate_fields(path, fns_keep, new_size_node = None):
    # fns_keep is the list of fields that need to keep
    # fns_keep needs to be in the same order as the data is stored in path
    # all fns_keep must be already in the path
    # fns_keep can be empty list
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    if new_size_node is None:
        ret = c.truncate_fields_sync_node(path, fns_keep)
    else:
        ret = c.truncate_fields_sync_node(path, fns_keep, new_size_node)
    if ret != 0:
        raise Exception(f"truncate_fields error {ret}")

def check_fields(path, is_check_all = True, new_size_node = None):
    # return list of field that is stored successful
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    is_only_check = True
    return properly_truncate_fields(path, is_check_all, is_only_check, new_size_node)

def check_compressed_eigen_vectors(path):
    # return bool value suggest whether the data can be read successfully
    # return True is the data has problem
    # return False if the data is ok
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    return c.check_compressed_eigen_vectors(path)

def eigen_system_repartition(new_size_node, path, path_new = ""):
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    if path_new[-14:] == "/geon-info.txt":
        path_new = path_new[:-14]
    return c.eigen_system_repartition(new_size_node, path, path_new)



import sys

usage_message = """
Usage:
    python3 -m qlat qlat-config ...
    python3 -m qlat eigen-system-checksum ...
    python3 -m qlat eigen-system-repartition ...
    python3 -m qlat fields-checksum ...
    python3 -m qlat fields-list ...
    python3 -m qlat fields-properly-truncate ...
    python3 -m qlat gauge-fix-coulomb ...
    python3 -m qlat topo-measure ...
""".strip()

if len(sys.argv) < 2:
    print(usage_message)
    exit()

action = sys.argv[1]
sys.argv = sys.argv[1:]

if action == "qlat-config":
    from .scripts import qlat_config
elif action == "eigen-system-checksum":
    from .scripts import eigen_system_checksum
elif action == "eigen-system-repartition":
    from .scripts import eigen_system_repartition
elif action == "fields-checksum":
    from .scripts import fields_checksum
elif action == "fields-list":
    from .scripts import fields_list
elif action == "fields-properly-truncate":
    from .scripts import fields_properly_truncate
elif action == "gauge-fix-coulomb":
    from .scripts import gauge_fix_coulomb
elif action == "topo-measure":
    from .scripts import topo_measure
else:
    print(usage_message)
    exit()



import sys
import qlat as q
import numpy as np

if len(sys.argv) == 1:
    q.displayln_info("Usage: topo-measure [ --source source_config ] [ --output output.pickle ]")

q.begin_with_mpi()

p_source = q.get_arg("--source")
p_output = q.get_arg("--output")

def load():
    if p_source is None:
        q.displayln_info("Need to provide source file with '--source filename'. Use a sample gauge field for now.")
        total_site = [ 4, 4, 4, 8, ]
        geo = q.Geometry(total_site, 1)
        gf = q.GaugeField(geo)
        rs = q.RngState("seed")
        gf.set_rand(rs.split("gf-init"), 0.5, 10)
    else:
        gf = q.GaugeField()
        gf.load(p_source)
    return gf

gf = load()

topo_list = q.smear_measure_topo(gf)

if p_output is not None:
    q.save_pickle_obj(topo_list, p_output)
else:
    q.displayln_info("To save the result, use '--output filename.pickle'. Print to screen for now.")
    q.displayln_info(q.pformat(topo_list))

q.timer_display()

q.end_with_mpi()

exit()



from qlat_utils import *
import qlat.c as c

from qlat.geometry import *
from qlat.field_selection import *
from qlat.field import *

from qlat.c import SelectedPoints, SelectedPointsBase

@timer
def set_selected_points(sp, f):
    # deprecated use @=
    from qlat.selected_field import SelectedField
    assert isinstance(sp, SelectedPointsBase)
    if isinstance(f, FieldBase):
        c.set_spfield_field(sp, f)
    elif isinstance(f, SelectedFieldBase):
        c.set_spfield_sfield(sp, f)
    else:
        raise Exception("set_selected_points")



"""
Qlattice main package.\n
Usage::\n
    import qlat as q\n
Will also load ``qlat_utils``.
"""

from qlat_utils import *

verbose_level(-1)

import qlat.c as c

from qlat.elem_type import *

from qlat.mpi import *

from qlat.geometry import *

from qlat.field import *

from qlat.field_utils import *

from qlat.utils import *

from qlat.utils_io import *

from qlat.qcd import *

from qlat.wilson_flow import *

from qlat.topology import *

from qlat.smear import *

from qlat.propagator import *

from qlat.gauge_action import *

from qlat.scalar_action import *

from qlat.hmc import *

from qlat.fthmc import *

from qlat.hmc_stats import *

from qlat.contract_pion import *

from qlat.contract_field import *

from qlat.field_selection import *

from qlat.selected_field import *

from qlat.selected_points import *

from qlat.fields_io import *

from qlat.inverter import *

from qlat.mat import *

import qlat.field_double

from qlat.get_include_dir import *

verbose_level("default")

del c



from .. import rbc_ukqcd_params as rup

job_tag = "96I"

dict_params = {}

rup.dict_params[job_tag] = dict_params

dict_params["job_tag"] = job_tag

dict_params["total_site"] = [ 96, 96, 96, 192, ]

dict_params["load_config_params"] = { "twist_boundary_at_boundary":[ 0.0, 0.0, 0.0, -0.5, ] }



from qlat_utils import *
import qlat.c as c

from qlat.field import *
from qlat.field_utils import *
from qlat.utils_io import *

class GaugeField(FieldColorMatrix):

    def __init__(self, geo = None):
        super().__init__(geo, 4)

    def copy(self, is_copying_data = True):
        f = GaugeField()
        if is_copying_data:
            f @= self
        return f

    @timer
    def save(self, path):
        """
        Save with the standard NERSC format
        """
        return c.save_gauge_field(self, path)

    @timer
    def load(self, path):
        """
        Load with the standard NERSC format
        """
        return c.load_gauge_field(self, path)

    def set_rand(self, rng, sigma = 0.5, n_step = 1):
        set_g_rand_color_matrix_field(self, rng, sigma, n_step)

    def unitarize(self):
        c.unitarize_color_matrix_field(self)

    def plaq(self):
        return gf_avg_plaq(self)

    def link_trace(self):
        return gf_avg_link_trace(self)

    def twist_boundary_at_boundary(self, lmom : float = -0.5, mu : int = 3):
        # modify in place
        gf_twist_boundary_at_boundary(self, lmom, mu)

    def show_info(self):
        gf_show_info(self)

###

class GaugeTransform(FieldColorMatrix):

    def __init__(self, geo = None):
        super().__init__(geo, 1)

    def copy(self, is_copying_data = True):
        f = GaugeTransform()
        if is_copying_data:
            f @= self
        return f

    @timer
    def save(self, path):
        """
        Save as double precision with the generic Field format ``save_double``
        """
        return self.save_double(path)

    @timer
    def load(self, path):
        """
        Load as double precision with the generic Field format ``load_double``
        """
        return self.load_double(path)

    @timer
    def save_cps(self, path):
        """
        Save with the format used in CPS
        """
        return c.save_gauge_transform_cps(self, path)

    @timer
    def load_cps(self, path):
        """
        Load with the format used in CPS
        """
        return c.load_gauge_transform_cps(self, path)

    def set_rand(self, rng, sigma = 0.5, n_step = 1):
        set_g_rand_color_matrix_field(self, rng, sigma, n_step)

    def unitarize(self):
        c.unitarize_color_matrix_field(self)

    def __mul__(self, other):
        # other can be GaugeTransform, GaugeField, Prop, SelProp, PselProp, list
        from qlat.propagator import Prop, SelProp, PselProp
        if isinstance(other, GaugeTransform):
            gt = GaugeTransform()
            c.apply_gt_gt(gt, self, other)
            return gt
        elif isinstance(other, GaugeField):
            gf = GaugeField()
            c.apply_gt_gf(gf, self, other)
            return gf
        elif isinstance(other, Prop):
            prop = Prop()
            c.apply_gt_prop(prop, self, other)
            return prop
        elif isinstance(other, SelProp):
            prop = SelProp(other.fsel)
            c.apply_gt_sprop(prop, self, other)
            return prop
        elif isinstance(other, PselProp):
            prop = PselProp(other.psel)
            c.apply_gt_psprop(prop, self, other)
            return prop
        elif isinstance(other, list):
            return [ self * p for p in other ]
        else:
            raise Exception("GaugeTransform.__mul__")

    def inv(self):
        gt = GaugeTransform()
        c.gt_invert(gt, self)
        return gt

###

@timer_verbose
def gf_show_info(gf):
    assert isinstance(gf, GaugeField)
    displayln_info(f"gf_show_info: plaq = {gf.plaq():.16F} ; link_trace = {gf.link_trace():.16F}.")

def gf_avg_plaq(gf):
    assert isinstance(gf, GaugeField)
    return c.gf_avg_plaq(gf)

def gf_avg_link_trace(gf):
    assert isinstance(gf, GaugeField)
    return c.gf_avg_link_trace(gf)

def gf_wilson_line_no_comm(wlf, m, gf_ext, path, path_n = None):
    # wlf = FieldColorMatrix(geo)
    # will only modify the m'th component
    # e.g. path = [ mu, mu, nu, -mu-1, -mu-1, ]
    # e.g. path = [ mu, nu, -mu-1, ], path_n = [ 2, 1, 2, ]
    if path_n is None:
        c.gf_wilson_line_no_comm(wlf, m, gf_ext, path)
    else:
        c.gf_wilson_line_no_comm(wlf, m, gf_ext, path, path_n)

def gf_wilson_lines_no_comm(gf_ext, path_list):
    # path_list = [ path_spec, ... ]
    # e.g. path_spec = [ mu, mu, nu, -mu-1, -mu-1, ]
    # e.g. path_spec = ([ mu, nu, -mu-1, ], [ 2, 1, 2, ],)
    # return wlf
    multiplicity = len(path_list)
    geo = geo_reform(gf_ext.geo(), multiplicity)
    wlf = FieldColorMatrix(geo)
    for m, p in enumerate(path_list):
        if isinstance(p, tuple) and len(p) == 2:
            path, path_n = p
            gf_wilson_line_no_comm(wlf, m, gf_ext, path, path_n)
        else:
            path = p
            gf_wilson_line_no_comm(wlf, m, gf_ext, path)
    return wlf

def gf_avg_wilson_loop_normalized_tr(gf, l, t):
    assert isinstance(gf, GaugeField)
    assert isinstance(l, int)
    assert isinstance(t, int)
    return c.gf_avg_wilson_loop_normalized_tr(gf, l, t)

def set_g_rand_color_matrix_field(fc, rng, sigma, n_steps = 1):
    assert isinstance(fc, FieldColorMatrix)
    assert isinstance(rng, RngState)
    return c.set_g_rand_color_matrix_field(fc, rng, sigma, n_steps)

def gf_twist_boundary_at_boundary(gf : GaugeField, lmom : float = -0.5, mu : int = 3):
    # modify gf in place
    c.gf_twist_boundary_at_boundary(gf, lmom, mu)

def mk_left_expanded_gauge_field(gf):
    gf1 = field_expanded(gf, 1, 0)
    refresh_expanded_1(gf1)
    return gf1



# Author: Luchang Jin 2022

import qlat as q
import sys

if len(sys.argv) not in [ 3, 4, ]:
    q.displayln_info("Usage: eigen-system-repartition 1.1.1.8 path [path_new]")
    q.displayln_info("Default path_new is the same as the old path.")
    exit()

q.begin_with_mpi()

new_size_node = [ int(x) for x in sys.argv[1].split(".") ]
path = sys.argv[2]
if len(sys.argv) == 4:
    path_new = sys.argv[3]
else:
    path_new = path

q.eigen_system_repartition(new_size_node, path, path_new)

q.timer_display()

q.end_with_mpi()

exit()



import qlat.c as c

from qlat.c import mod_coordinate
from qlat.c import smod_coordinate

from qlat_utils import index_from_coordinate, coordinate_from_index

import numpy as np

def mk_epsilon_array():
    arr = np.zeros((4, 4, 4, 4,), dtype = np.int8)
    def setv(i, j, k, l):
        setw(i, j, k, l, 1)
        setw(i, j, l, k, -1)
    def setw(i, j, k, l, val):
        arr[i, j, k, l] = val;
        arr[j, k, l, i] = -val;
        arr[k, l, i, j] = val;
        arr[l, i, j, k] = -val;
    setv(0, 1, 2, 3)
    setv(0, 2, 3, 1);
    setv(0, 3, 1, 2);
    return arr

epsilon_array = mk_epsilon_array()

def epsilon_tensor(i, j, k, l = 3):
    return epsilon_array[i, j, k, l].item()



from . import rbc_ukqcd as ru
from . import rbc_ukqcd_params as rup
from .jobs import *

# -----------------------------------------------------------------------------

@q.timer
def run_get_inverter(job_tag, traj, *, inv_type, get_gf, get_gt = None, get_eig = None):
    if None in [ get_gf, ]:
        return
    if get_gt is None:
        get_gt = lambda: None
    if get_eig is None:
        get_eig = lambda: None
    gf = get_gf()
    gt = get_gt()
    eig = get_eig()
    for inv_acc in [ 0, 1, 2, ]:
        ru.get_inv(gf, job_tag, inv_type, inv_acc, gt = gt, eig = eig)

# -----------------------------------------------------------------------------

@q.timer_verbose
def compute_prop_1(inv, src, *, tag, sfw, path_sp, psel, fsel, fselc):
    fn_sp = os.path.join(path_sp, f"{tag}.lat")
    fn_spw = os.path.join(path_sp, f"{tag} ; wsnk.lat")
    sol = inv * src
    s_sol = q.SelProp(fselc)
    s_sol @= sol
    s_sol.save_float_from_double(sfw, tag)
    sp_sol = q.PselProp(psel)
    sp_sol @= s_sol
    sp_sol.save(get_save_path(fn_sp))
    sfw.flush()
    sol_ws = sol.glb_sum_tslice()
    sol_ws.save(get_save_path(fn_spw))
    return sol

@q.timer
def compute_prop_wsrc(gf, gt, tslice, job_tag, inv_type, inv_acc, *,
        idx, sfw, path_sp, psel, fsel, fselc, eig, finished_tags):
    tag = f"tslice={tslice} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in finished_tags:
        return None
    q.check_stop()
    q.check_time_limit()
    q.displayln_info(f"compute_prop_wsrc: idx={idx} tslice={tslice}", job_tag, inv_type, inv_acc)
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, gt = gt, eig = eig)
    total_site = ru.get_total_site(job_tag)
    geo = q.Geometry(total_site, 1)
    src = q.mk_wall_src(geo, tslice)
    prop = compute_prop_1(inv, src, tag = tag, sfw = sfw, path_sp = path_sp,
                          psel = psel, fsel = fsel, fselc = fselc)

@q.timer_verbose
def compute_prop_wsrc_all(job_tag, traj, *,
                          inv_type, gf, gt, wi, psel, fsel, fselc, eig):
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    path_s = f"{job_tag}/prop-wsrc-{inv_type_name}/traj-{traj}"
    path_sp = f"{job_tag}/psel-prop-wsrc-{inv_type_name}/traj-{traj}"
    finished_tags = q.properly_truncate_fields(get_save_path(path_s + ".acc"))
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", [ 1, 2, 2, 4, ])
    for inv_acc in [ 2, 1 ]:
        for p in wi:
            idx, tslice, inv_type_p, inv_acc_p = p
            if inv_type_p == inv_type and inv_acc_p == inv_acc:
                compute_prop_wsrc(gf, gt, tslice, job_tag, inv_type, inv_acc,
                        idx = idx, sfw = sfw, path_sp = path_sp,
                        psel = psel, fsel = fsel, fselc = fselc, eig = eig,
                        finished_tags = finished_tags)
    sfw.close()
    q.qtouch_info(get_save_path(os.path.join(path_sp, "checkpoint.txt")))
    # q.qtouch_info(get_save_path(os.path.join(path_sp, "checkpoint ; wsnk.txt")))
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
    q.qar_create_info(get_save_path(path_sp + ".qar"), get_save_path(path_sp), is_remove_folder_after = True)
    # q.qar_create_info(get_save_path(path_s + ".qar"), get_save_path(path_s), is_remove_folder_after = True)

@q.timer
def run_prop_wsrc(job_tag, traj, *, inv_type, get_gf, get_eig, get_gt, get_psel, get_fsel, get_wi):
    if None in [ get_gf, get_gt, get_psel, get_fsel, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    if get_load_path(f"{job_tag}/prop-wsrc-{inv_type_name}/traj-{traj}/geon-info.txt") is not None:
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-wsrc-{inv_type_name}"):
        gf = get_gf()
        gt = get_gt()
        eig = get_eig()
        fsel, fselc = get_fsel()
        psel = get_psel()
        wi = get_wi()
        compute_prop_wsrc_all(job_tag, traj,
                              inv_type = inv_type, gf = gf, gt = gt, wi = wi,
                              psel = psel, fsel = fsel, fselc = fselc, eig = eig)
        q.release_lock()

# -----------------------------------------------------------------------------

@q.timer_verbose
def compute_prop_2(inv, src, *, tag, sfw, path_sp, psel, fsel, fselc, gt):
    fn_sp = os.path.join(path_sp, f"{tag}.lat")
    fn_spw = os.path.join(path_sp, f"{tag} ; wsnk.lat")
    sol = inv * src
    s_sol = q.SelProp(fselc)
    s_sol @= sol
    s_sol.save_float_from_double(sfw, tag)
    sp_sol = q.PselProp(psel)
    sp_sol @= s_sol
    sp_sol.save(get_save_path(fn_sp))
    sfw.flush()
    sol_gt = gt * sol
    sol_ws = sol_gt.glb_sum_tslice()
    sol_ws.save(get_save_path(fn_spw))
    return sol

@q.timer
def compute_prop_psrc(job_tag, xg_src, inv_type, inv_acc, *,
        idx, gf, gt, sfw, path_sp, psel, fsel, fselc, eig, finished_tags):
    xg = xg_src
    xg_str = f"({xg[0]},{xg[1]},{xg[2]},{xg[3]})"
    tag = f"xg={xg_str} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in finished_tags:
        return None
    q.check_stop()
    q.check_time_limit()
    q.displayln_info(f"compute_prop_psrc: {job_tag} idx={idx} tag='{tag}'")
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, eig = eig)
    total_site = get_param(job_tag, "total_site")
    geo = q.Geometry(total_site, 1)
    src = q.mk_point_src(geo, xg_src)
    prop = compute_prop_2(inv, src, tag = tag, sfw = sfw, path_sp = path_sp,
                          psel = psel, fsel = fsel, fselc = fselc, gt = gt)

@q.timer_verbose
def compute_prop_psrc_all(job_tag, traj, *,
                          inv_type, gf, gt, psel, fsel, fselc, eig):
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    path_s = f"{job_tag}/prop-psrc-{inv_type_name}/traj-{traj}"
    path_sp = f"{job_tag}/psel-prop-psrc-{inv_type_name}/traj-{traj}"
    finished_tags = q.properly_truncate_fields(get_save_path(path_s + ".acc"))
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", [ 1, 2, 2, 4, ])
    def comp(idx, xg_src, inv_acc):
        compute_prop_psrc(job_tag, xg_src, inv_type, inv_acc,
                idx = idx, gf = gf, gt = gt, sfw = sfw, path_sp = path_sp,
                psel = psel, fsel = fsel, fselc = fselc,
                eig = eig, finished_tags = finished_tags)
    prob1 = get_param(job_tag, "prob_acc_1_psrc")
    prob2 = get_param(job_tag, "prob_acc_2_psrc")
    rs = q.RngState(f"seed {job_tag} {traj}").split(f"compute_prop_psrc_all(ama)")
    for idx, xg_src in enumerate(psel.to_list()):
        r = rs.split(f"{tuple(xg_src)}").u_rand_gen()
        assert 0 <= r and r <= 1
        comp(idx, xg_src, inv_acc = 0)
        if r <= prob1:
            comp(idx, xg_src, inv_acc = 1)
        if r <= prob2:
            comp(idx, xg_src, inv_acc = 2)
    sfw.close()
    q.qtouch_info(get_save_path(os.path.join(path_sp, "checkpoint.txt")))
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
    q.qar_create_info(get_save_path(path_sp + ".qar"), get_save_path(path_sp), is_remove_folder_after = True)
    # q.qar_create_info(get_save_path(path_s + ".qar"), get_save_path(path_s), is_remove_folder_after = True)

@q.timer
def run_prop_psrc(job_tag, traj, *, inv_type, get_gf, get_eig, get_gt, get_psel, get_fsel):
    if None in [ get_gf, get_gt, get_psel, get_fsel, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    if get_load_path(f"{job_tag}/prop-psrc-{inv_type_name}/traj-{traj}/geon-info.txt") is not None:
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-psrc-{inv_type_name}"):
        gf = get_gf()
        gt = get_gt()
        eig = get_eig()
        fsel, fselc = get_fsel()
        psel = get_psel()
        compute_prop_psrc_all(job_tag, traj,
                              inv_type = inv_type, gf = gf, gt = gt,
                              psel = psel, fsel = fsel, fselc = fselc, eig = eig)
        q.release_lock()

# -----------------------------------------------------------------------------

@q.timer_verbose
def compute_prop_rand_u1_type_acc(*, sfw, job_tag, traj, gf, eig, fsel, idx_rand_u1, inv_type, inv_acc, finished_tags):
    # same rand source for different inv_type
    tag = f"idx_rand_u1={idx_rand_u1} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in finished_tags:
        return
    q.check_stop()
    q.check_time_limit()
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, eig = eig)
    rs = q.RngState(f"seed {job_tag} {traj}").split(f"compute_prop_rand_u1(rand_u1)").split(str(idx_rand_u1))
    s_prop = q.mk_rand_u1_prop(inv, fsel, rs)
    s_prop.save_float_from_double(sfw, tag)
    sfw.flush()
    return s_prop

@q.timer_verbose
def compute_prop_rand_u1(*, job_tag, traj, inv_type, gf, path_s, fsel, eig = None):
    # use fsel instead of fselc
    n_rand_u1_fsel = get_param(job_tag, "n_rand_u1_fsel")
    total_site = get_param(job_tag, "total_site")
    geo = q.Geometry(total_site, 1)
    finished_tags = q.properly_truncate_fields(get_save_path(path_s + ".acc"))
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", [ 1, 2, 2, 4, ])
    def comp(idx_rand_u1, inv_acc):
        compute_prop_rand_u1_type_acc(
                sfw = sfw,
                job_tag = job_tag, traj = traj,
                gf = gf, eig = eig, fsel = fsel,
                idx_rand_u1 = idx_rand_u1,
                inv_type = inv_type, inv_acc = inv_acc,
                finished_tags = finished_tags,)
    rs = q.RngState(f"seed {job_tag} {traj}").split(f"compute_prop_rand_u1(ama)")
    prob1 = get_param(job_tag, "prob_acc_1_rand_u1")
    prob2 = get_param(job_tag, "prob_acc_2_rand_u1")
    for idx_rand_u1 in range(n_rand_u1_fsel):
        r = rs.split(str(idx_rand_u1)).u_rand_gen()
        inv_acc = 0
        assert 0 <= r and r <= 1
        comp(idx_rand_u1, inv_acc)
        inv_acc = 1
        if r <= prob1:
            comp(idx_rand_u1, inv_acc)
        inv_acc = 2
        if r <= prob2:
            comp(idx_rand_u1, inv_acc)
    sfw.close()
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
    # q.qar_create_info(get_save_path(path_s + ".qar"), get_save_path(path_s), is_remove_folder_after = True)

@q.timer_verbose
def run_prop_rand_u1(job_tag, traj, *, inv_type, get_gf, get_fsel, get_eig = None):
    if None in [ get_gf, get_fsel, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_names = [ "light", "strange", "charm", ]
    inv_type_name = inv_type_names[inv_type]
    path_s = f"{job_tag}/prop-rand-u1-{inv_type_name}/traj-{traj}"
    if get_load_path(path_s + "/geon-info.txt") is not None:
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-rand-u1-{inv_type_name}"):
        gf = get_gf()
        fsel, fselc = get_fsel()
        eig = get_eig()
        compute_prop_rand_u1(
                job_tag = job_tag, traj = traj,
                inv_type = inv_type,
                gf = gf,
                path_s = path_s,
                fsel = fsel,
                eig = eig)
        q.release_lock()

# -----------------------------------------------------------------------------

@q.timer_verbose
def compute_prop_3(inv, src_smear, *, tag, sfw, path_sp, psel, fsel, fselc, gt, psel_smear, smear):
    fn_sp = os.path.join(path_sp, f"{tag}.lat")
    fn_spw = os.path.join(path_sp, f"{tag} ; wsnk.lat")
    fn_sps = os.path.join(path_sp, f"{tag} ; smear-snk.lat")
    sol = inv * src_smear
    s_sol = q.SelProp(fselc)
    s_sol @= sol
    s_sol.save_float_from_double(sfw, tag)
    sp_sol = q.PselProp(psel)
    sp_sol @= s_sol
    sp_sol.save(get_save_path(fn_sp))
    sfw.flush()
    sol_gt = gt * sol
    sol_ws = sol_gt.glb_sum_tslice()
    sol_ws.save(get_save_path(fn_spw))
    sol_smear = sol.copy()
    smear(sol_smear)
    sol_smear_psel = q.PselProp(psel_smear)
    sol_smear_psel @= sol_smear
    sol_smear_psel.save(get_save_path(fn_sps))
    return sol

@q.timer
def compute_prop_smear(job_tag, xg_src, inv_type, inv_acc, *,
        idx, gf, gt, sfw, path_sp, psel, fsel, fselc, psel_smear, gf_ape, eig, finished_tags):
    xg = xg_src
    xg_str = f"({xg[0]},{xg[1]},{xg[2]},{xg[3]})"
    tag = f"smear ; xg={xg_str} ; type={inv_type} ; accuracy={inv_acc}"
    if tag in finished_tags:
        return None
    q.check_stop()
    q.check_time_limit()
    q.displayln_info(f"compute_prop_smear: {job_tag} idx={idx} tag='{tag}'")
    inv = ru.get_inv(gf, job_tag, inv_type, inv_acc, eig = eig)
    total_site = ru.get_total_site(job_tag)
    geo = q.Geometry(total_site, 1)
    coef = get_param(job_tag, "prop_smear_coef")
    step = get_param(job_tag, "prop_smear_step")
    def smear(src):
        q.prop_smear(src, gf_ape, coef, step)
    src = q.mk_point_src(geo, xg_src)
    smear(src)
    prop = compute_prop_3(inv, src, tag = tag, sfw = sfw, path_sp = path_sp,
                          psel = psel, fsel = fsel, fselc = fselc, gt = gt,
                          psel_smear = psel_smear, smear = smear)

@q.timer_verbose
def compute_prop_smear_all(job_tag, traj, *,
        inv_type, gf, gt, psel, fsel, fselc, psel_smear, gf_ape, eig,
        ):
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    path_s = f"{job_tag}/prop-smear-{inv_type_name}/traj-{traj}"
    path_sp = f"{job_tag}/psel-prop-smear-{inv_type_name}/traj-{traj}"
    finished_tags = q.properly_truncate_fields(get_save_path(path_s + ".acc"))
    sfw = q.open_fields(get_save_path(path_s + ".acc"), "a", [ 1, 2, 2, 4, ])
    def comp(idx, xg_src, inv_acc):
        compute_prop_smear(job_tag, xg_src, inv_type, inv_acc,
                idx = idx, gf = gf, gt = gt, sfw = sfw, path_sp = path_sp,
                psel = psel, fsel = fsel, fselc = fselc,
                psel_smear = psel_smear, gf_ape = gf_ape,
                eig = eig, finished_tags = finished_tags)
    prob1 = get_param(job_tag, "prob_acc_1_smear")
    prob2 = get_param(job_tag, "prob_acc_2_smear")
    rs = q.RngState(f"seed {job_tag} {traj}").split(f"compute_prop_smear_all(ama)")
    for idx, xg_src in enumerate(psel_smear.to_list()):
        r = rs.split(f"{tuple(xg_src)}").u_rand_gen()
        assert 0 <= r and r <= 1
        comp(idx, xg_src, inv_acc = 0)
        if r <= prob1:
            comp(idx, xg_src, inv_acc = 1)
        if r <= prob2:
            comp(idx, xg_src, inv_acc = 2)
    sfw.close()
    q.qtouch_info(get_save_path(os.path.join(path_sp, "checkpoint.txt")))
    q.qrename_info(get_save_path(path_s + ".acc"), get_save_path(path_s))
    q.qar_create_info(get_save_path(path_sp + ".qar"), get_save_path(path_sp), is_remove_folder_after = True)
    # q.qar_create_info(get_save_path(path_s + ".qar"), get_save_path(path_s), is_remove_folder_after = True)

@q.timer
def run_prop_smear(job_tag, traj, *, inv_type, get_gf, get_gf_ape, get_eig, get_gt, get_psel, get_fsel, get_psel_smear):
    if None in [ get_gf, get_gt, get_gf_ape, get_psel, get_fsel, ]:
        return
    if get_eig is None:
        if inv_type == 0:
            return
        get_eig = lambda: None
    inv_type_names = [ "light", "strange", ]
    inv_type_name = inv_type_names[inv_type]
    if get_load_path(f"{job_tag}/prop-smear-{inv_type_name}/traj-{traj}/geon-info.txt") is not None:
        return
    if q.obtain_lock(f"locks/{job_tag}-{traj}-smear-{inv_type_name}"):
        gf = get_gf()
        gt = get_gt()
        eig = get_eig()
        psel = get_psel()
        fsel, fselc = get_fsel()
        psel_smear = get_psel_smear()
        gf_ape = get_gf_ape()
        compute_prop_smear_all(job_tag, traj,
                inv_type = inv_type, gf = gf, gf_ape = gf_ape, gt = gt,
                psel = psel, fsel = fsel, fselc = fselc, eig = eig, psel_smear = psel_smear)
        q.release_lock()

# -----------------------------------------------------------------------------



# all cpp level import

from qlat_utils.everything cimport *

cdef extern from "qlat/mpi.h" namespace "qlat":

    int begin(const int id_node, const Coordinate& size_node, const int color) except +
    int end(const bool is_preserving_cache) except +

cdef extern from "qlat/geometry.h" namespace "qlat":

    cdef cppclass Geometry:
        int multiplicity
        Geometry()
        const Geometry& operator=(const Geometry& geo) except +
        void init()
        void init(Coordinate& total_site, int multiplicity) except +
        Coordinate total_site()
        Coordinate local_site()
        Coordinate local_volume()
        Coordinate total_volume()
    std_string show(const Geometry& geo) except +
    Geometry geo_resize(const Geometry& geo, int thick) except +
    Geometry geo_resize(const Geometry& geo,
                        const Coordinate& expansion_left, const Coordinate& expansion_right) except +
    Geometry geo_reform(const Geometry& geo, int multiplicity, int thick) except +
    Geometry geo_reform(const Geometry& geo, int multiplicity,
                        const Coordinate& expansion_left, const Coordinate& expansion_right) except +
    Geometry geo_eo(const Geometry& geo, int eo) except +

cdef extern from "qlat/utils-io.h" namespace "qlat":

    int qremove_info(const std_string& path) except +
    int qremove_all_info(const std_string& path) except +

cdef extern from "qlat/field.h" namespace "qlat":

    cdef cppclass Field[T]:
        Field()
        void init()
        void init(const Geometry& geo) except +
        void init(const Geometry& geo, int multiplicity) except +
        void init(const Field[T]& field) except +
        const Field[T]& operator=(const Field[T]& field) except +
        const Geometry& get_geo()
    Vector[T] get_data[T](const Field[T]& x)
    void set_zero[T](Field[T]& x)
    void qswap[T](Field[T]& x, Field[T]& y) except +

cdef extern from "qlat/selected-field-io.h" namespace "qlat":

    cdef cppclass FieldSelection:
        FieldSelection()
        const Geometry& get_geo()
    cdef cppclass PointsSelection:
        PointsSelection()
        PointsSelection(const long n_points) except +
        long size()
        Coordinate* data()
        Coordinate& operator[](long i)
    cdef cppclass SelectedField[T]:
        long n_elems;
        SelectedField()
        void init()
        void init(const Geometry& geo, const long n_elems, const int multiplicity) except +
        void init(const FieldSelection& fsel, const int multiplicity) except +
        const SelectedField[T]& operator=(const SelectedField[T]& field) except +
        const Geometry& get_geo()
    cdef cppclass SelectedPoints[T]:
        int multiplicity
        long n_points
        SelectedPoints()
        void init()
        void init(const long n_points, const int multiplicity) except +
        void init(const PointsSelection& psel, const int multiplicity) except +
        const SelectedPoints[T]& operator=(const SelectedPoints[T]& field) except +
    Vector[T] get_data[T](const SelectedField[T]& x)
    void set_zero[T](SelectedField[T]& x)
    void qswap[T](SelectedField[T]& x, SelectedField[T]& y)
    Vector[T] get_data[T](const SelectedPoints[T]& x)
    void set_zero[T](SelectedPoints[T]& x)
    void qswap[T](SelectedPoints[T]& x, SelectedPoints[T]& y) except +



cdef class FieldTYPENAME(FieldBase):

    cdef cqlat.Field[cqlat.TYPENAME] xx

cdef class SelectedFieldTYPENAME(SelectedFieldBase):

    cdef cqlat.SelectedField[cqlat.TYPENAME] xx

cdef class SelectedPointsTYPENAME(SelectedPointsBase):

    cdef cqlat.SelectedPoints[cqlat.TYPENAME] xx
